import weakref
from os import path
from PyQt5 import uic
from pyqtgraph.Qt import QtCore, QtGui
from linien.client.gui import ui_path


class IDSelector:
    def __init__(self, parent):
        self.parent = parent

    def __getattr__(self, name):
        return self.parent.get_widget(name)


class CustomWidget:
    instances = []

    def __init__(self, *args, **kwargs):
        self.__class__.instances.append(weakref.proxy(self))

        super().__init__(*args, **kwargs)

        self.ids = IDSelector(self)

        QtCore.QTimer.singleShot(100, self.ready)

    def ready(self):
        pass

    def connection_established(self):
        pass

    def get_widget(self, name):
        """Queries a widget by name."""
        return self.findChild(QtCore.QObject, name)

    def app(self):
        # this property is set manually. Probably there is a more elegant way
        # to solve this...
        return self.window().app

    def load_ui(self, name):
        assert name.endswith('.ui')
        uic.loadUi(path.join(ui_path, name), self)


import os
import paramiko
from plumbum import colors
from PyQt5.QtWidgets import QSlider, QCheckBox, QSpinBox, QDoubleSpinBox, \
    QTabWidget, QRadioButton, QComboBox

import linien
from linien.config import REMOTE_BASE_PATH
from linien.client.exceptions import InvalidServerVersionException, \
    ServerNotInstalledException


def connect_ssh(host, user, password):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostname=host, username=user, password=password)
    return ssh


def run_server(host, user, password):
    ssh = connect_ssh(host, user, password)

    version = linien.__version__

    if version == 'dev':
        # if we are in a development version, we upload the files from source
        # directory directly to the RP.
        upload_source_code(ssh)

        # start the development code that was just uploaded.
        # For that, go to the parent directory; an import of "linien" then
        # points to that directory instead of any globally installed release
        # version.
        stdin, stdout, stderr = ssh.exec_command(
            ('cd %s/../;' % REMOTE_BASE_PATH) +
            ('bash %s/server/linien_start_server' % REMOTE_BASE_PATH)
        )
        err = stderr.read()
        if err:
            print(colors.red | 'Error starting the server')
            print(err)
    else:
        # it is a release and not a dev version.

        # read out version number of the server
        remote_version = read_remote_version(ssh)
        print('remote version', remote_version, 'local version', version)

        if version != remote_version:
            raise InvalidServerVersionException(version, remote_version)

        # start the server process using the global command
        ssh.exec_command('linien_start_server')
        ssh.close()


def read_remote_version(ssh):
    """Reads the remote version of linien using SSH."""
    stdin, stdout, stderr = ssh.exec_command(
        'python3 -c "import linien; print(linien.__version__);"'
    )
    lines = stdout.readlines()

    if not lines:
        raise ServerNotInstalledException()

    remote_version = lines[0].strip()
    return remote_version


def upload_source_code(ssh):
    """Uploads the application's source code to the remote server using SFTP."""
    print('uploading dev source code..')

    ftp = ssh.open_sftp()

    directory = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '..'
        )
    )

    # upload the code required for running the server
    for dirpath, dirnames, filenames in os.walk(directory):
        dirpath_rel = dirpath.replace(directory, '').lstrip('/')
        remote_path = os.path.join(REMOTE_BASE_PATH, dirpath_rel)

        if '.' in dirpath_rel or '__' in dirpath_rel:
            continue

        try:
            ftp.lstat(remote_path)
        except IOError:
            ftp.mkdir(os.path.join(remote_path.rstrip('/')))

        for filename in filenames:
            local_path = os.path.join(dirpath, filename)
            remote_filepath = os.path.join(remote_path, filename)
            # put file
            ftp.put(local_path, remote_filepath)

    ftp.close()


def param2ui(parameter, element, process_value=lambda x: x):
    """Updates ui elements according to parameter values.

    Listens to parameter changes and sets the value of `element` automatically.
    Optionally, the value can be processed using `process_value`.
    This function should be used because it automatically blocks signal
    emission from the target element; otherwise this can cause nasty
    endless loops when quickly changing a paramater multiple times.
    """
    def on_change(value, element=element):
        element.blockSignals(True)

        value = process_value(value)

        if isinstance(element, (QSlider, QSpinBox, QDoubleSpinBox)):
            element.setValue(value)
        elif isinstance(element, (QCheckBox, QRadioButton)):
            element.setChecked(value)
        elif isinstance(element, (QTabWidget, QComboBox)):
            element.setCurrentIndex(int(value))
        else:
            raise Exception('unsupported element type %s' % type(element))

        element.blockSignals(False)

    parameter.change(on_change)


from PyQt5 import QtGui
from linien.client.widgets import CustomWidget
from linien.client.utils import param2ui


class LockingPanel(QtGui.QWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_ui('locking_panel.ui')

    def ready(self):
        self.ids.kp.setKeyboardTracking(False)
        self.ids.kp.valueChanged.connect(self.kp_changed)
        self.ids.ki.setKeyboardTracking(False)
        self.ids.ki.valueChanged.connect(self.ki_changed)
        self.ids.kd.setKeyboardTracking(False)
        self.ids.kd.valueChanged.connect(self.kd_changed)
        self.ids.watchLockCheckbox.stateChanged.connect(self.watch_lock_changed)
        self.ids.lock_control_container.currentChanged.connect(self.lock_mode_changed)

        self.ids.manualLockButton.clicked.connect(self.start_manual_lock)
        self.ids.autoOffsetCheckbox.stateChanged.connect(self.auto_offset_changed)

        self.ids.pid_on_slow_strength.setKeyboardTracking(False)
        self.ids.pid_on_slow_strength.valueChanged.connect(self.pid_on_slow_strength_changed)

    def connection_established(self):
        params = self.app().parameters
        self.parameters = params
        self.control = self.app().control

        param2ui(params.p, self.ids.kp)
        param2ui(params.i, self.ids.ki)
        param2ui(params.d, self.ids.kd)

        param2ui(params.watch_lock, self.ids.watchLockCheckbox)
        param2ui(params.autolock_determine_offset, self.ids.autoOffsetCheckbox)
        param2ui(
            params.automatic_mode,
            self.ids.lock_control_container,
            lambda value: 0 if value else 1
        )
        param2ui(params.pid_on_slow_strength, self.ids.pid_on_slow_strength)
        def slow_pid_visibility(*args):
            self.ids.slow_pid_group.setVisible(self.parameters.pid_on_slow_enabled.value)
        params.pid_on_slow_enabled.change(slow_pid_visibility)

        def lock_status_changed(_):
            locked = params.lock.value
            task = params.task.value
            al_failed = params.autolock_failed.value
            task_running = (task is not None) and (not al_failed)

            if locked or task_running:
                self.ids.lock_control_container.hide()
            else:
                self.ids.lock_control_container.show()

        for param in (params.lock, params.autolock_approaching, params.autolock_watching,
                      params.autolock_failed, params.autolock_locked):
            param.change(lock_status_changed)

        param2ui(params.target_slope_rising, self.ids.button_slope_rising)
        param2ui(
            params.target_slope_rising,
            self.ids.button_slope_falling,
            lambda value: not value
        )

    def kp_changed(self):
        self.parameters.p.value = self.ids.kp.value()
        self.control.write_data()

    def ki_changed(self):
        self.parameters.i.value = self.ids.ki.value()
        self.control.write_data()

    def kd_changed(self):
        self.parameters.d.value = self.ids.kd.value()
        self.control.write_data()

    def watch_lock_changed(self):
        self.parameters.watch_lock.value = int(self.ids.watchLockCheckbox.checkState())

    def lock_mode_changed(self, idx):
        self.parameters.automatic_mode.value = idx == 0

    def start_manual_lock(self):
        self.parameters.target_slope_rising.value = self.ids.button_slope_rising.isChecked()
        self.control.write_data()
        self.control.start_lock()

    def auto_offset_changed(self):
        self.parameters.autolock_determine_offset.value = int(self.ids.autoOffsetCheckbox.checkState())

    def pid_on_slow_strength_changed(self):
        self.parameters.pid_on_slow_strength.value = self.ids.pid_on_slow_strength.value()
        self.control.write_data()


import os
import sys
sys.path += ['../../']
import click
import _thread
import pickle

import rpyc
from rpyc.utils.server import ThreadedServer

from autolock import Autolock
from parameters import Parameters

from linien.config import SERVER_PORT
from linien.common import update_control_signal_history
from linien.communication.server import BaseService


class RedPitayaControlService(BaseService):
    """Control server that runs on the RP that provides high-level methods."""
    def __init__(self, **kwargs):
        self._cached_data = {}
        self.exposed_is_locked = None

        super().__init__(Parameters)

        from registers import Registers
        self.registers = Registers(**kwargs)
        self.registers.connect(self, self.parameters)

        self._skip_next_data = 0

    def run_acquiry_loop(self):
        """Starts a background process that keeps polling control and error
        signal. Every received value is pushed to `parameters.to_plot`."""
        def data_received(plot_data):
            # When a parameter is changed, `pause_acquisition` is set.
            # This means that the we should skip new data until we are sure that
            # it was recorded with the new settings.
            if not self.parameters.pause_acquisition.value:
                if self._skip_next_data > 1:
                    self._skip_next_data -= 1
                else:
                    s1, s2, slow_out = pickle.loads(plot_data)
                    is_locked = self.parameters.lock.value

                    if is_locked:
                        data = {
                            'error_signal': s1,
                            'control_signal': s2
                        }
                        if self.parameters.pid_on_slow_enabled.value:
                            data['slow'] = slow_out
                    else:
                        data = {
                            'error_signal_1': s1,
                            'error_signal_2': s2
                        }

                    self.parameters.to_plot.value = pickle.dumps(data)

                    self.parameters.control_signal_history.value = \
                        update_control_signal_history(
                            self.parameters.control_signal_history.value,
                            data,
                            is_locked,
                            self.parameters.control_signal_history_length.value
                        )

        self.registers.run_data_acquisition(data_received)

    def exposed_write_data(self):
        """Syncs the parameters with the FPGA registers."""
        self.registers.write_registers()

    def exposed_start_autolock(self, x0, x1, spectrum, auto_offset=True):
        spectrum = pickle.loads(spectrum)
        start_watching = self.parameters.watch_lock.value
        auto_offset = self.parameters.autolock_determine_offset.value

        if not self.parameters.autolock_running.value:
            autolock = Autolock(self, self.parameters)
            self.parameters.task.value = autolock
            autolock.run(x0, x1, spectrum, should_watch_lock=start_watching,
                         auto_offset=auto_offset)

    def exposed_start_ramp(self):
        self.pause_acquisition()

        self.parameters.combined_offset.value = 0
        self.parameters.lock.value = False
        self.exposed_write_data()

        self.continue_acquisition()

    def exposed_start_lock(self):
        self.pause_acquisition()

        self.parameters.lock.value = True
        self.exposed_write_data()

        self.continue_acquisition()

    def exposed_reset_scan(self):
        """Resets scan mode."""
        self.pause_acquisition()

        self.parameters.ramp_amplitude.value = 1
        self.parameters.center.value = 0
        self.exposed_start_ramp()
        self.exposed_write_data()

        self.continue_acquisition()

    def exposed_shutdown(self):
        """Kills the server."""
        self.registers.acquisition.shutdown()
        _thread.interrupt_main()
        os._exit(0)

    def exposed_get_server_version(self):
        import linien
        return linien.__version__

    def exposed_get_restorable_parameters(self):
        return self.parameters.restorable_parameters

    def pause_acquisition(self):
        """Pause continuous acquisition. Call this before changing a parameter
        that alters the error / control signal. This way, no inconsistent signals
        reach the application. After setting the new parameter values, call
        `continue_acquisition`."""
        self.parameters.pause_acquisition.value = True

    def continue_acquisition(self):
        """Continus acquisition after a short delay, when we are sure that the
        new parameters values have been written to the FPGA and that data that
        is now recorded is recorded with the correct parameters."""
        self.parameters.pause_acquisition.value = False
        self._skip_next_data = 3


class FakeRedPitayaControl(BaseService):
    def __init__(self):
        super().__init__(Parameters)
        self.exposed_is_locked = None

    def exposed_write_data(self):
        pass

    def run_acquiry_loop(self):
        import threading
        from time import sleep
        from random import randint

        def run():
            while True:
                max_ = randint(0, 8191)
                self.parameters.to_plot.value = pickle.dumps((
                    [randint(-max_, max_) for _ in range(16384)],
                    list(_ - 8192 for _ in range(16384))
                ))
                sleep(.1)
        t = threading.Thread(target=run)
        t.daemon = True
        t.start()

    def exposed_shutdown(self):
        _thread.interrupt_main()
        os._exit(0)

    def exposed_start_autolock(self, x0, x1, spectrum):
        print('start autolock', x0, x1)

    def exposed_get_restorable_parameters(self):
        return self.parameters.restorable_parameters

    def exposed_get_server_version(self):
        import linien
        return linien.__version__

    def pause_acquisition(self):
        pass

    def continue_acquisition(self):
        pass


@click.command()
@click.option('--fake', is_flag=True,
              help='Runs a fake server that just returns random data')
@click.option('--remote-rp',
              help='Allows to run the server locally for development and '
                   'connects to a RedPitaya. Specify the RP\'s credentials '
                   'as follows: '
                   '--remote-rp=root:myPassword@rp-f0xxxx.local')
def run_server(fake, remote_rp):
    if fake:
        print('starting fake server')
        control = FakeRedPitayaControl()
    else:
        if remote_rp is not None:
            assert '@' in remote_rp and ':' in remote_rp, 'invalid format, should be root:myPassword@rp-f0xxxx.local'

            username, tmp = remote_rp.split(':', 1)
            r_host, r_password = ''.join(reversed(tmp)).split('@', 1)
            host = ''.join(reversed(r_host))
            password = ''.join(reversed(r_password))
            control = RedPitayaControlService(
                host=host, user=username, password=password
            )
        else:
            control = RedPitayaControlService()

    control.run_acquiry_loop()
    control.exposed_write_data()

    t = ThreadedServer(control, port=SERVER_PORT)
    t.start()


if __name__ == '__main__':
    run_server()


import numpy as np
from PyQt5 import QtGui

from linien.common import convert_channel_mixing_value, FAST_OUT1, FAST_OUT2, \
    ANALOG_OUT0
from linien.client.utils import param2ui
from linien.client.widgets import CustomWidget
from linien.client.connection import MHz, Vpp


class GeneralPanel(QtGui.QWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def ready(self):
        self.ids.channel_mixing_slider.valueChanged.connect(self.channel_mixing_changed)
        self.ids.dual_channel.stateChanged.connect(self.dual_channel_changed)

        self.ids.mod_channel.currentIndexChanged.connect(self.mod_channel_changed)
        self.ids.control_channel.currentIndexChanged.connect(self.control_channel_changed)
        self.ids.sweep_channel.currentIndexChanged.connect(self.sweep_channel_changed)
        self.ids.slow_control_channel.currentIndexChanged.connect(self.slow_control_channel_changed)

        self.ids.polarity_fast_out1.currentIndexChanged.connect(self.polarity_fast_out1_changed)
        self.ids.polarity_fast_out2.currentIndexChanged.connect(self.polarity_fast_out2_changed)
        self.ids.polarity_analog_out0.currentIndexChanged.connect(self.polarity_analog_out0_changed)

    def connection_established(self):
        params = self.app().parameters
        self.control = self.app().control
        self.parameters = params

        def dual_channel_changed(value):
            self.ids.dual_channel_mixing.setVisible(value)
            self.app().main_window.ids.spectroscopy_channel_2_page.setEnabled(value)
            return value
        param2ui(
            params.dual_channel,
            self.ids.dual_channel,
            dual_channel_changed
        )

        param2ui(
            params.channel_mixing,
            self.ids.channel_mixing_slider,
            lambda value: value + 128
        )
        # this is required to update the descriptive labels in the beginning
        self.channel_mixing_changed()

        param2ui(params.mod_channel, self.ids.mod_channel)
        param2ui(params.control_channel, self.ids.control_channel)
        param2ui(params.sweep_channel, self.ids.sweep_channel)
        param2ui(params.pid_on_slow_enabled, self.ids.slow_control_channel)

        param2ui(params.polarity_fast_out1, self.ids.polarity_fast_out1)
        param2ui(params.polarity_fast_out2, self.ids.polarity_fast_out2)
        param2ui(params.polarity_analog_out0, self.ids.polarity_analog_out0)

        def show_polarity_settings(*args):
            used_channels = set((
                params.control_channel.value,
                params.sweep_channel.value,
            ))

            if params.pid_on_slow_enabled.value:
                used_channels.add(ANALOG_OUT0)

            self.ids.polarity_selector.setVisible(len(used_channels) > 1)

            def set_visibility(element, channel_id):
                element.setVisible(channel_id in used_channels)

            set_visibility(self.ids.polarity_container_fast_out1, FAST_OUT1)
            set_visibility(self.ids.polarity_container_fast_out2, FAST_OUT2)
            set_visibility(self.ids.polarity_container_analog_out0, ANALOG_OUT0)
        params.control_channel.change(show_polarity_settings)
        params.sweep_channel.change(show_polarity_settings)
        params.mod_channel.change(show_polarity_settings)
        params.pid_on_slow_enabled.change(show_polarity_settings)

    def channel_mixing_changed(self):
        value = int(self.ids.channel_mixing_slider.value()) - 128
        self.parameters.channel_mixing.value = value
        self.control.write_data()

        self.update_channel_mixing_slider(value)

    def dual_channel_changed(self):
        self.parameters.dual_channel.value = int(self.ids.dual_channel.checkState() > 0)
        self.control.write_data()

    def update_channel_mixing_slider(self, value):
        a_value, b_value = convert_channel_mixing_value(value)

        self.ids.chain_a_factor.setText('%d' % a_value)
        self.ids.chain_b_factor.setText('%d' % b_value)

    def mod_channel_changed(self, channel):
        self.parameters.mod_channel.value = channel
        self.control.write_data()

    def control_channel_changed(self, channel):
        self.parameters.control_channel.value = channel
        self.control.write_data()

    def slow_control_channel_changed(self, channel):
        self.parameters.pid_on_slow_enabled.value = bool(channel)
        self.control.write_data()

    def sweep_channel_changed(self, channel):
        self.parameters.sweep_channel.value = channel
        self.control.write_data()

    def polarity_fast_out1_changed(self, polarity):
        self.parameters.polarity_fast_out1.value = bool(polarity)
        self.control.write_data()

    def polarity_fast_out2_changed(self, polarity):
        self.parameters.polarity_fast_out2.value = bool(polarity)
        self.control.write_data()

    def polarity_analog_out0_changed(self, polarity):
        self.parameters.polarity_analog_out0.value = bool(polarity)
        self.control.write_data()


import sys
sys.path += ['../../']
import rpyc
import atexit
import threading

from enum import Enum
from time import sleep
from multiprocessing import Process, Pipe

from linien.config import ACQUISITION_PORT
from linien.server.utils import stop_nginx, start_nginx, flash_fpga

class AcquisitionConnectionError(Exception):
    pass


class AcquisitionProcessSignals(Enum):
    SHUTDOWN = 0
    SET_RAMP_SPEED = 2
    SET_LOCK_STATUS = 3
    SET_CSR = 4
    SET_IIR_CSR = 5


class AcquisitionMaster:
    def __init__(self, use_ssh, host):
        self.on_acquisition = None

        def receive_acquired_data(conn):
            while True:
                received = conn.recv()
                if self.on_acquisition is not None:
                    self.on_acquisition(received)

        self.acq_process, child_pipe = Pipe()
        p = Process(
            target=self.connect_acquisition_process,
            args=(child_pipe, use_ssh, host)
        )
        p.daemon = True
        p.start()

        # wait until connection is established
        self.acq_process.recv()

        t = threading.Thread(target=receive_acquired_data, args=(self.acq_process,))
        t.daemon = True
        t.start()

        atexit.register(self.shutdown)

    def run_data_acquisition(self, on_acquisition):
        self.on_acquisition = on_acquisition

    def connect_acquisition_process(self, pipe, use_ssh, host):
        if use_ssh:
            acquisition_rpyc = rpyc.connect(host, ACQUISITION_PORT)
            acquisition = acquisition_rpyc.root
        else:
            from linien.server.acquisition_process import DataAcquisitionService
            stop_nginx()
            flash_fpga()
            acquisition = DataAcquisitionService()

        # tell the main thread that we're ready
        pipe.send(True)

        # run a loop that listens for acquired data and transmits them
        # to the main thread. Also redirects calls from the main thread
        # to the acquiry process.
        last_hash = None
        while True:
            # check whether the main thread sent a command to the acquiry process
            while pipe.poll():
                data = pipe.recv()
                if data[0] == AcquisitionProcessSignals.SHUTDOWN:
                    raise SystemExit()
                elif data[0] == AcquisitionProcessSignals.SET_RAMP_SPEED:
                    speed = data[1]
                    acquisition.exposed_set_ramp_speed(speed)
                elif data[0] == AcquisitionProcessSignals.SET_LOCK_STATUS:
                    acquisition.exposed_set_lock_status(data[1])
                elif data[0] == AcquisitionProcessSignals.SET_CSR:
                    acquisition.exposed_set_csr(*data[1])
                elif data[0] == AcquisitionProcessSignals.SET_IIR_CSR:
                    acquisition.exposed_set_iir_csr(*data[1])

            # load acquired data and send it to the main thread
            current_hash = acquisition.exposed_get_data_hash()

            if current_hash != last_hash:
                last_hash = current_hash

                data = acquisition.exposed_return_data()
                pipe.send(data)

            sleep(0.05)

    def shutdown(self):
        if self.acq_process:
            self.acq_process.send((AcquisitionProcessSignals.SHUTDOWN,))

        start_nginx()

    def set_ramp_speed(self, speed):
        self.acq_process.send((AcquisitionProcessSignals.SET_RAMP_SPEED, speed))

    def lock_status_changed(self, status):
        if self.acq_process:
            self.acq_process.send((AcquisitionProcessSignals.SET_LOCK_STATUS, status))

    def set_csr(self, key, value):
        self.acq_process.send(
            (AcquisitionProcessSignals.SET_CSR, (key, value))
        )

    def set_iir_csr(self, *args):
        self.acq_process.send(
            (AcquisitionProcessSignals.SET_IIR_CSR, args)
        )


import shutil
import subprocess

from linien.config import REMOTE_BASE_PATH


def stop_nginx():
    subprocess.Popen(['systemctl', 'stop', 'redpitaya_nginx.service']).wait()
    subprocess.Popen(['systemctl', 'stop', 'redpitaya_scpi.service']).wait()


def start_nginx():
    subprocess.Popen(['systemctl', 'start', 'redpitaya_nginx.service'])


def flash_fpga():
    shutil.copyfile('redpid.bin', '/dev/xdevcfg')


def twos_complement(num, N_bits):
    max_ = 1<<(N_bits - 1)
    full = 2 * max_

    if num < 0:
        num += full

    return num



"""
    gain_camera.communication.server
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Utils for storing and managing a set of parameters on the server side.
    `gain_camera.communication.client` contains utils for accessing it.
"""
import rpyc
from time import sleep

from .utils import pack, unpack


class Parameter:
    """Represents a single parameter and is used by `Parameters`."""
    def __init__(self, min_=None, max_=None, start=None, wrap=False, sync=True,
                 collapsed_sync=False):
        self.min = min_
        self.max = max_
        self.wrap = wrap
        self._value = start
        self._start = start
        self._listeners = set()
        self._collapsed_sync = collapsed_sync
        self.exposed_sync = sync

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        # check bounds
        if self.min is not None and value < self.min:
            value = self.min if not self.wrap else self.max
        if self.max is not None and value > self.max:
            value = self.max if not self.wrap else self.min

        self._value = value

        # we copy it because a listener could remove a listener --> this would
        # cause an error in this loop
        for listener in self._listeners.copy():
            listener(value)

    def change(self, function):
        self._listeners.add(function)

        if self._value is not None:
            function(self._value)

    def remove_listener(self, function):
        if function in self._listeners:
            self._listeners.remove(function)

    def exposed_reset(self):
        self.value = self._start

    def register_remote_listener(self, remote_uuid):
        pass


class BaseParameters:
    """Represents a set of parameters. In an actual program, it should be
    sub-classed like this:

        class MyParameters(BaseParameters):
            def __init__(self):
                self.param1 = Parameter(min_=12, max_=24)

    Parameters can be changed like this:

        p = MyParameters(...)
        p.param1.value = 123

    You can register callback functions like this:

        def on_change(value):
            # do something

        p.param1.change(on_change)
    """
    def __init__(self):
        self._remote_listener_queue = {}
        self._remote_listener_callbacks = {}

    def get_all_parameters(self):
        for name, element in self.__dict__.items():
            if isinstance(element, Parameter):
                yield name, element

    def register_remote_listener(self, uuid, param_name):
        self._remote_listener_queue.setdefault(uuid, [])
        self._remote_listener_callbacks.setdefault(uuid, [])

        def on_change(value, uuid=uuid, param_name=param_name):
            if uuid in self._remote_listener_queue:
                self._remote_listener_queue[uuid].append((param_name, value))

        param = getattr(self, param_name)
        param.change(on_change)

        self._remote_listener_callbacks[uuid].append((param, on_change))

    def unregister_remote_listeners(self, uuid):
        for param, callback in self._remote_listener_callbacks[uuid]:
            param.remove_listener(callback)

        del self._remote_listener_queue[uuid]
        del self._remote_listener_callbacks[uuid]

    def get_listener_queue(self, uuid):
        queue = self._remote_listener_queue.get(uuid, [])
        self._remote_listener_queue[uuid] = []

        # filter out multiple values for collapsible parameters
        already_has_value = []
        for idx in reversed(range(len(queue))):
            param_name, value = queue[idx]
            if self._get_param(param_name)._collapsed_sync:
                if param_name in already_has_value:
                    del queue[idx]
                else:
                    already_has_value.append(param_name)

        return pack(queue)

    def __iter__(self):
        for name, param in self.get_all_parameters():
            yield name, param.value

    def _get_param(self, param_name):
        param = getattr(self, param_name)
        assert isinstance(param, Parameter)
        return param


class BaseService(rpyc.Service):
    """A service that provides functionality for seamless integration of
    parameter access on the client."""
    def __init__(self, parameter_cls):
        self.parameters = parameter_cls()
        self._uuid_mapping = {}

    def on_connect(self, client):
        self._uuid_mapping[client] = client.root.uuid

    def on_disconnect(self, client):
        uuid = self._uuid_mapping[client]
        self.parameters.unregister_remote_listeners(uuid)

    def exposed_get_param(self, param_name):
        return pack(self.parameters._get_param(param_name).value)

    def exposed_set_param(self, param_name, value):
        self.parameters._get_param(param_name).value = unpack(value)

    def exposed_get_all_parameters(self):
        return self.parameters.get_all_parameters()

    def exposed_register_remote_listener(self, uuid, param_name):
        return self.parameters.register_remote_listener(uuid, param_name)

    def exposed_get_listener_queue(self, uuid):
        return self.parameters.get_listener_queue(uuid)


from PyQt5 import QtGui, QtWidgets
from linien.common import LOW_PASS_FILTER, HIGH_PASS_FILTER
from linien.client.widgets import CustomWidget
from linien.client.utils import param2ui


class SpectroscopyPanel(QtWidgets.QWidget, CustomWidget):
    def __init__(self, *args):
        super().__init__(*args)
        self.load_ui('spectroscopy_panel.ui')

        def change_filter_frequency(filter_i):
            self.get_param('filter_%d_frequency' % filter_i).value = \
                getattr(self.ids, 'filter_%d_frequency' % filter_i).value()
            self.control.write_data()

        def change_filter_enabled(filter_i):
            filter_enabled = int(
                getattr(self.ids, 'filter_%d_enabled' % filter_i).checkState() > 0
            )
            self.get_param('filter_%d_enabled' % filter_i).value = filter_enabled
            self.control.write_data()

        def change_filter_type(filter_idx):
            param = self.get_param('filter_%d_type' % filter_i)
            current_idx = getattr(self.ids, 'filter_%d_type' % filter_i).currentIndex()
            param.value = (LOW_PASS_FILTER, HIGH_PASS_FILTER)[current_idx]
            self.control.write_data()

        for filter_i in [1, 2]:
            for key, fct in {
                'change_filter_%d_frequency': change_filter_frequency,
                'change_filter_%d_enabled': change_filter_enabled,
                'change_filter_%d_type': change_filter_type
            }.items():
                setattr(self, key % filter_i, lambda *args, fct=fct, filter_i=filter_i: fct(filter_i))

    def get_param(self, name):
        return getattr(self.parameters, '%s_%s' % (name, 'a' if self.channel == 0 else 'b'))

    def ready(self):
        self.ids.signal_offset.setKeyboardTracking(False)
        self.ids.signal_offset.valueChanged.connect(self.change_signal_offset)
        self.ids.demodulation_phase.setKeyboardTracking(False)
        self.ids.demodulation_phase.valueChanged.connect(self.change_demod_phase)
        self.ids.demodulation_frequency.currentIndexChanged.connect(self.change_demod_multiplier)

        for filter_i in [1, 2]:
            _get = lambda parent, attr, filter_i=filter_i: getattr(parent, attr % filter_i)
            _get(self.ids, 'filter_%d_enabled') \
                .stateChanged.connect(_get(self, 'change_filter_%d_enabled'))
            freq_input = _get(self.ids, 'filter_%d_frequency')
            freq_input.setKeyboardTracking(False)
            freq_input.valueChanged.connect(_get(self, 'change_filter_%d_frequency'))
            _get(self.ids, 'filter_%d_type') \
                .currentIndexChanged.connect(_get(self, 'change_filter_%d_type'))

    def connection_established(self):
        params = self.app().parameters
        self.control = self.app().control
        self.parameters = params

        #self.close_button.clicked.connect(self.close_app)
        #self.shutdown_button.clicked.connect(self.shutdown_server)

        param2ui(
            self.get_param('demodulation_phase'),
            self.ids.demodulation_phase
        )
        param2ui(
            self.get_param('demodulation_multiplier'),
            self.ids.demodulation_frequency,
            lambda value: value - 1
        )
        param2ui(
            self.get_param('offset'),
            self.ids.signal_offset
        )
        for filter_i in [1, 2]:
            param2ui(
                self.get_param('filter_%d_enabled' % filter_i),
                getattr(self.ids, 'filter_%d_enabled' % filter_i)
            )
            param2ui(
                self.get_param('filter_%d_frequency' % filter_i),
                getattr(self.ids, 'filter_%d_frequency' % filter_i)
            )
            param2ui(
                self.get_param('filter_%d_type' % filter_i),
                getattr(self.ids, 'filter_%d_type' % filter_i),
                lambda type_: {
                    LOW_PASS_FILTER: 0,
                    HIGH_PASS_FILTER: 1
                }[type_]
            )

    def change_signal_offset(self):
        self.get_param('offset').value = self.ids.signal_offset.value()
        self.control.write_data()

    def change_demod_phase(self):
        self.get_param('demodulation_phase').value = self.ids.demodulation_phase.value()
        self.control.write_data()

    def change_demod_multiplier(self, idx):
        self.get_param('demodulation_multiplier').value = idx + 1
        self.control.write_data()


class SpectroscopyChannelAPanel(SpectroscopyPanel):
    channel = 0


class SpectroscopyChannelBPanel(SpectroscopyPanel):
    channel = 1


from .utils import unpack, pack


class RemoteParameter:
    """A helper class for `RemoteParameters`, representing a single remote
    parameter."""
    def __init__(self, parent, remote, name, use_cache):
        self.remote = remote
        self.name = name
        self.parent = parent

        if use_cache and self.remote.exposed_sync:
            self.change(self._sync_value)

    @property
    def value(self):
        if hasattr(self, '_cached_value'):
            return self._cached_value
        return self.parent._get_param(self.name)

    @value.setter
    def value(self, value):
        return self.parent._set_param(self.name, value)

    def change(self, function):
        self.parent.register_listener(self, function)
        function(self.value)

    def reset(self):
        self.remote.reset()

    @property
    def _start(self):
        return self.remote._start

    def _sync_value(self, value):
        self._cached_value = value


class RemoteParameters:
    """A class that provides remote access to a `parameters.Parameters` instance.

    It clones the functionality of the remote `Parameters` instance. E.g.:

        r = RemoteParameters(...)
        r.my_param.value = 123

        def on_change(value):
            # do something

        r.my_param.change(on_change)
    """
    def __init__(self, remote, uuid, use_cache):
        self.remote = remote
        self.uuid = uuid

        self._listeners = {}

        for name, param in remote.exposed_get_all_parameters():
            setattr(self, name, RemoteParameter(self, param, name, use_cache))

        self.call_listeners()

    def __iter__(self):
        for name, param in self.remote.exposed_get_all_parameters():
            yield name, getattr(self, name).value

    def register_listener(self, param, callback):
        if param.name not in self._listeners:
            self.remote.exposed_register_remote_listener(self.uuid, param.name)

        self._listeners.setdefault(param.name, [])
        self._listeners[param.name].append(callback)

    def call_listeners(self):
        queue = unpack(self.remote.get_listener_queue(self.uuid))

        for param_name, value in queue:
            for listener in self._listeners[param_name]:
                listener(value)

    def _get_param(self, param_name):
        return unpack(self.remote.exposed_get_param(param_name))

    def _set_param(self, param_name, value):
        return self.remote.exposed_set_param(param_name, pack(value))



import math
import pickle
import numpy as np
import pyqtgraph as pg

from time import time
from PyQt5 import QtGui, QtWidgets
from pyqtgraph.Qt import QtCore, QtGui

from linien.client.config import COLORS
from linien.client.widgets import CustomWidget
from linien.common import update_control_signal_history, determine_shift_by_correlation, \
    get_lock_point, control_signal_has_correct_amplitude, combine_error_signal, \
    check_plot_data


class PlotWidget(pg.PlotWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.hideAxis('bottom')
        #self.hideAxis('left')

        self.setMouseEnabled(x=False, y=False)
        self.setMenuEnabled(False)

        # important: increasing pen width makes plotting much slower!
        # alternative: pg.setConfigOptions(useOpenGL=True)
        # see: https://github.com/pyqtgraph/pyqtgraph/issues/533
        pen_width = 1

        self.zero_line = pg.PlotCurveItem(pen=pg.mkPen('w', width=1))
        self.addItem(self.zero_line)
        self.signal1 = pg.PlotCurveItem(pen=pg.mkPen(COLORS['spectroscopy1'], width=pen_width))
        self.addItem(self.signal1)
        self.signal2 = pg.PlotCurveItem(pen=pg.mkPen(COLORS['spectroscopy2'], width=pen_width))
        self.addItem(self.signal2)
        self.combined_signal = pg.PlotCurveItem(pen=pg.mkPen(COLORS['spectroscopy_combined'], width=pen_width))
        self.addItem(self.combined_signal)

        self.control_signal = pg.PlotCurveItem(pen=pg.mkPen(COLORS['control_signal'], width=pen_width))
        self.addItem(self.control_signal)
        self.control_signal_history = pg.PlotCurveItem(pen=pg.mkPen(COLORS['control_signal_history'], width=pen_width))
        self.addItem(self.control_signal_history)
        self.slow_history = pg.PlotCurveItem(pen=pg.mkPen(COLORS['slow_history'], width=pen_width))
        self.addItem(self.slow_history)

        self.zero_line.setData([0, 16383], [0, 0])
        self.signal1.setData([0, 16383], [1, 1])
        self.signal1.setData([0, 16383], [1, 1])
        self.combined_signal.setData([0, 16383], [1, 1])

        self.connection = None
        self.parameters = None
        self.last_plot_rescale = 0
        self.last_plot_data = None
        self.plot_max = 0
        self.plot_min = np.inf
        self.touch_start = None
        self.autolock_ref_spectrum = None

        self.init_overlay()
        self.init_lock_target_line()

    def _to_data_coords(self, event):
        pos = self.plotItem.vb.mapSceneToView(event.pos())
        x, y = pos.x(), pos.y()
        return x, y

    def connection_established(self):
        self.control = self.app().control
        self.parameters = self.app().parameters

        self.control_signal_history_data = self.parameters.control_signal_history.value

        self.parameters.to_plot.change(self.replot)

    def mouseMoveEvent(self, event):
        if self.touch_start is None:
            return

        x0, y0 = self.touch_start

        x, y = self._to_data_coords(event)
        self.set_selection_overlay(x0, x - x0)

    def init_overlay(self):
        self.overlay = pg.LinearRegionItem(values=(-1000, -1000), movable=False)
        self.overlay.setVisible(False)
        self.addItem(self.overlay)

    def init_lock_target_line(self):
        self.lock_target_line = pg.InfiniteLine(movable=False)
        self.lock_target_line.setValue(1000)
        self.addItem(self.lock_target_line)

    def set_selection_overlay(self, x_start, width):
        self.overlay.setRegion((x_start, x_start + width))

    def mousePressEvent(self, event):
        super().mousePressEvent(event)

        if event.button() == QtCore.Qt.RightButton:
            return

        x, y = self._to_data_coords(event)

        self.touch_start = x, y
        self.set_selection_overlay(x, 0)
        self.overlay.setVisible(True)

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)

        if self.touch_start is None:
            return

        x, y = self._to_data_coords(event)
        x0, y0 = self.touch_start
        xdiff = np.abs(x0 - x)

        if xdiff / self.xmax < 0.01:
            # it was a click
            if not self.parameters.automatic_mode.value:
                self.graph_on_click(x0, y0)
        else:
            # it was a selection
            if self.parameters.automatic_mode.value:
                last_combined_error_signal = self.last_plot_data[2]
                self.control.start_autolock(
                    *sorted([x0, x]),
                    # we pickle it here because otherwise a netref is
                    # transmitted which blocks the autolock
                    pickle.dumps(last_combined_error_signal)
                )
                mean_signal, target_slope_rising, target_zoom, rolled_error_signal = \
                    get_lock_point(last_combined_error_signal, int(x0), int(x))
                self.autolock_ref_spectrum = rolled_error_signal
            else:
                self.graph_on_selection(x0, x)

        self.overlay.setVisible(False)
        self.touch_start = None

    @property
    def xmax(self):
        return len(self.last_plot_data[0]) - 1

    def graph_on_selection(self, x0, x):
        x0 /= self.xmax
        x /= self.xmax

        center = np.mean([x, x0])
        amplitude = np.abs(center - x) * 2
        center = (center - 0.5) * 2

        amplitude *= self.parameters.ramp_amplitude.value
        center = self.parameters.center.value + \
            (center * self.parameters.ramp_amplitude.value)

        self.parameters.ramp_amplitude.value = amplitude
        self.parameters.center.value = center
        self.control.write_data()

    def graph_on_click(self, x, y):
        center = x / self.xmax
        center = (center - 0.5) * 2
        center = self.parameters.center.value + \
            (center * self.parameters.ramp_amplitude.value)

        self.parameters.ramp_amplitude.value /= 2
        self.parameters.center.value = center
        self.control.write_data()

    def replot(self, to_plot):
        if self.parameters.pause_acquisition.value:
            return

        if to_plot is not None and not self.touch_start:
            to_plot = pickle.loads(to_plot)

            if to_plot is None:
                return

            if not check_plot_data(self.parameters.lock.value, to_plot):
                return

            # we also call this if the laser is not locked because it resets
            # the history in this case
            history, slow_history = self.update_control_signal_history(to_plot)

            if self.parameters.lock.value:
                self.last_plot_data = to_plot

                self.signal1.setVisible(False)
                self.signal2.setVisible(False)
                self.control_signal.setVisible(True)
                self.control_signal_history.setVisible(True)
                self.slow_history.setVisible(self.parameters.pid_on_slow_enabled.value)
                self.combined_signal.setVisible(True)

                error_signal, control_signal = to_plot['error_signal'], to_plot['control_signal']
                all_signals = (error_signal, control_signal, history, slow_history)

                self.plot_data_locked(to_plot)
                self.update_plot_scaling(all_signals)
                self.plot_autolock_target_line(None)
            else:
                self.signal1.setVisible(True)
                self.signal2.setVisible(self.parameters.dual_channel.value)
                self.combined_signal.setVisible(self.parameters.dual_channel.value)
                self.control_signal.setVisible(False)
                self.control_signal_history.setVisible(False)
                self.slow_history.setVisible(False)

                s1, s2 = to_plot['error_signal_1'], to_plot['error_signal_2']
                combined_error_signal = combine_error_signal(
                    (s1, s2),
                    self.parameters.dual_channel.value,
                    self.parameters.channel_mixing.value
                )
                all_signals = [s1, s2] + [combined_error_signal]
                self.last_plot_data = all_signals

                self.plot_data_unlocked((s1, s2), combined_error_signal)
                self.plot_autolock_target_line(combined_error_signal)
                self.update_plot_scaling(all_signals)

    def plot_data_unlocked(self, error_signals, combined_signal):
        error_signal1, error_signal2 = error_signals
        self.signal1.setData(list(range(len(error_signal1))), error_signal1)
        self.signal2.setData(list(range(len(error_signal2))), error_signal2)
        self.combined_signal.setData(list(range(len(combined_signal))), combined_signal)

    def plot_data_locked(self, signals):
        error_signal = signals['error_signal']
        control_signal = signals['control_signal']
        self.combined_signal.setData(list(range(len(error_signal))), error_signal)
        self.control_signal.setData(list(range(len(error_signal))), control_signal)

    def plot_autolock_target_line(self, combined_error_signal):
        if self.autolock_ref_spectrum is not None and self.parameters.autolock_approaching.value:
            ramp_amplitude = self.parameters.ramp_amplitude.value
            zoom_factor = 1 / ramp_amplitude
            initial_zoom_factor = 1 / self.parameters.autolock_initial_ramp_amplitude.value

            shift, _1, _2 = determine_shift_by_correlation(
                zoom_factor / initial_zoom_factor,
                self.autolock_ref_spectrum,
                combined_error_signal
            )
            shift *= zoom_factor / initial_zoom_factor
            length = len(combined_error_signal)
            shift = (length / 2) - (shift / 2* length)

            self.lock_target_line.setVisible(True)
            self.lock_target_line.setValue(shift)
        else:
            self.lock_target_line.setVisible(False)

    def update_plot_scaling(self, signals):
        if time() - self.last_plot_rescale > .5:
            all_ = []
            for signal in signals:
                all_ += signal

            if self.parameters.autoscale_y.value:
                self.plot_min = math.floor(np.min(all_))
                self.plot_max = math.ceil(np.max(all_))
            else:
                self.plot_min, self.plot_max = self.parameters.y_axis_limits.value

            if self.plot_min == self.plot_max:
                self.plot_max += 1

            self.setYRange(self.plot_min, self.plot_max)

            self.last_plot_rescale = time()

    def update_control_signal_history(self, to_plot):
        self.control_history_data = update_control_signal_history(
            self.control_signal_history_data,
            to_plot,
            self.parameters.lock.value,
            self.parameters.control_signal_history_length.value
        )
        if self.parameters.lock.value:
            x_axis_length = 16384

            def scale(arr):
                timescale = self.parameters.control_signal_history_length.value
                if arr:
                    arr = np.array(arr)
                    arr -= arr[0]
                    arr *= 1 / timescale * x_axis_length
                return arr

            history = self.control_signal_history_data['values']
            self.control_signal_history.setData(
                scale(self.control_signal_history_data['times']),
                history
            )

            slow_values = self.control_signal_history_data['slow_values']
            self.slow_history.setData(
                scale(self.control_signal_history_data['slow_times']),
                slow_values
            )

            return history, slow_values
        return [], []


from linien.communication.server import Parameter, BaseParameters


class Parameters(BaseParameters):
    def __init__(self):
        super().__init__()

        # parameters whose values are saved on the client and restored if no
        # server is running
        self.restorable_parameters = (
            'modulation_amplitude', 'modulation_frequency', 'ramp_speed',
            'demodulation_phase_a', 'demodulation_multiplier_a',
            'demodulation_phase_b', 'demodulation_multiplier_b',
            'offset_a', 'offset_b',
            'filter_1_enabled_a', 'filter_1_enabled_b',
            'filter_1_frequency_a', 'filter_1_frequency_b',
            'filter_1_type_a', 'filter_1_type_b',
            'filter_2_enabled_a', 'filter_2_enabled_b',
            'filter_2_frequency_a', 'filter_2_frequency_b',
            'filter_2_type_a', 'filter_2_type_b',
            'p', 'i', 'd', 'watch_lock', 'dual_channel',
            'channel_mixing',
            'pid_on_slow_enabled', 'pid_on_slow_strength',
            'mod_channel', 'control_channel', 'sweep_channel',
            'polarity_fast_out1', 'polarity_fast_out2',
            'polarity_analog_out0', 'autoscale_y', 'y_axis_limits'
        )

        self.modulation_amplitude = Parameter(
            min_=0,
            max_=(1<<14) - 1,
            start=4046
        )
        self.modulation_frequency = Parameter(
            min_=0,
            max_=0xffffffff,
            # 0x10000000 ~= 8 MHzs
            start=0x10000000/8*15
        )
        self.center = Parameter(
            min_=-1,
            max_=1,
            start=0
        )

        self.ramp_amplitude = Parameter(
            min_=0.001,
            max_=1,
            start=1
        )
        self.ramp_speed = Parameter(
            min_=0,
            max_=16,
            start=9
        )

        for channel in ('a', 'b'):
            setattr(self, 'demodulation_phase_%s' % channel, Parameter(
                min_=0,
                max_=360,
                start=0x0,
                wrap=True
            ))
            setattr(self, 'demodulation_multiplier_%s' % channel, Parameter(
                min_=0,
                max_=15,
                start=1
            ))
            setattr(self, 'offset_%s' % channel, Parameter(
                min_=-8191,
                max_=8191,
                start=0
            ))
            for filter_i in [1, 2]:
                setattr(self, 'filter_%d_enabled_%s' % (filter_i, channel), Parameter(start=False))
                setattr(self, 'filter_%d_type_%s' % (filter_i, channel), Parameter(start=0))
                setattr(self, 'filter_%d_frequency_%s' % (filter_i, channel), Parameter(start=10000))

        self.combined_offset = Parameter(
            min_=-8191,
            max_=8191,
            start=0
        )

        self.lock = Parameter(start=False)
        self.to_plot = Parameter()

        self.p = Parameter(start=50)
        self.i = Parameter(start=5)
        self.d = Parameter(start=0)
        self.task = Parameter(start=None, sync=False)
        self.automatic_mode = Parameter(start=True)
        self.target_slope_rising = Parameter(start=True)
        self.autolock_running = Parameter(start=False)
        self.autolock_approaching = Parameter(start=False)
        self.autolock_watching = Parameter(start=False)
        self.autolock_failed = Parameter(start=False)
        self.autolock_locked = Parameter(start=False)
        self.autolock_determine_offset = Parameter(start=True)
        self.autolock_initial_ramp_amplitude = Parameter(start=1)
        self.pause_acquisition = Parameter(start=False)

        self.watch_lock = Parameter(start=True)
        self.control_signal_history = Parameter(start={
            'times': [],
            'values': []
        }, sync=False)
        # in seconds
        self.control_signal_history_length = Parameter(start=600)

        self.pid_on_slow_enabled = Parameter(start=False)
        self.pid_on_slow_strength = Parameter(start=0)
        self.dual_channel = Parameter(start=False)
        self.channel_mixing = Parameter(start=0)

        self.mod_channel = Parameter(start=0, min_=0, max_=1)
        self.control_channel = Parameter(start=1, min_=0, max_=1)
        self.sweep_channel = Parameter(start=1, min_=0, max_=2)

        self.polarity_fast_out1 = Parameter(start=False)
        self.polarity_fast_out2 = Parameter(start=False)
        self.polarity_analog_out0 = Parameter(start=False)

        self.autoscale_y = Parameter(start=True)
        self.y_axis_limits = Parameter(start=((-1000, 1000)))


import bz2
import pickle


def pack(value):
    try:
        return pickle.dumps(value)
    except (AttributeError, TypeError):
        # this happens when un-pickleable objects (e.g. functions) are assigned
        # to a parameter. In this case, we don't pickle it but transfer a netref
        # instead
        return value


def unpack(value):
    try:
        return pickle.loads(value)
    except:
        return value


# Copyright 2014-2015 Robert Jordens <jordens@gmail.com>
#
# This file is part of redpid.
#
# redpid is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# redpid is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with redpid.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import subprocess

import csrmap
from iir_coeffs import make_filter, get_params


class PitayaCSR:
    map = csrmap.csr
    constants = csrmap.csr_constants
    offset = 0x40300000

    def set(self, name, value):
        map, addr, width, wr = self.map[name]
        assert wr, name

        ma = 1 << width
        bit_mask = ma - 1
        val = value & bit_mask
        assert value == val or ma + value == val, (
            'value for %s out of range' % name, (value, val, ma)
        )

        b = (width + 8 - 1) // 8
        for i in range(b):
            v = (val >> (8*(b - i - 1))) & 0xff
            self.set_one(self.offset + (map << 11) + ((addr + i)<<2), v)

    def get(self, name):
        if name in self.constants:
            return self.constants[name]

        map, addr, nr, wr = self.map[name]
        v = 0
        b = (nr + 8 - 1)//8
        for i in range(b):
            v |= self.get_one(self.offset + (map << 11) + ((addr + i)<<2)
                    ) << 8*(b - i - 1)
        return v

    def set_iir(self, prefix, b, a, z=0):
        shift = self.get(prefix + "_shift") or 16
        width = self.get(prefix + "_width") or 18
        interval = self.get(prefix + "_interval") or 1
        b, a, params = get_params(b, a, shift, width, interval)

        for k in sorted(params):
            self.set(prefix + "_" + k, params[k])
        self.set(prefix + "_z0", z)
        for i in range(len(b), 3):
            n = prefix + "_b%i" % i
            if n in self.map:
                self.set(n, 0)
                self.set(prefix + "_a%i" % i, 0)

    def signal(self, name):
        return csrmap.signals.index(name)

    def states(self, *names):
        return sum(1<<csrmap.states.index(name) for name in names)


class PitayaSSH(PitayaCSR):
    def __init__(self, ssh_cmd='ssh root@rp-f0685a.local', monitor_cmd="/usr/bin/monitoradvanced"):
        self.p = subprocess.Popen(' '.join([ssh_cmd, monitor_cmd, '-']).split(),
                stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)

    def set_one(self, addr, value):
        cmd = "0x{:08x} w 0x{:02x}\n\n".format(addr, value)
        self.p.stdin.write(cmd.encode("ascii"))
        self.p.stdin.flush()

    def get_one(self, addr):
        cmd = "0x{:08x} w\n\n".format(addr)
        self.p.stdin.write(cmd.encode("ascii"))
        self.p.stdin.flush()
        ret = self.p.stdout.readline().decode("ascii")
        return int(ret.split('[10]')[-1].strip())


class PitayaLocal(PitayaCSR):
    def __init__(self, monitor_cmd="/usr/bin/monitoradvanced"):
        self.monitor_cmd = monitor_cmd
        self.start_process()

    def start_process(self):
        self.command_counter = 0
        self.p = subprocess.Popen([self.monitor_cmd, '-'],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def check_for_process_restart(self):
        """monitoradvanced has a memory leak. Therefore, we restart it from time
        to time (not elegant but works)."""
        if self.command_counter > 1000:
            self.command_counter = 0
            self.p.kill()
            self.start_process()

    def set_one(self, addr, value):
        self.command_counter += 1
        self.check_for_process_restart()
        cmd = "0x{:08x} w 0x{:02x}\n\n".format(addr, value)
        self.p.stdin.write(cmd.encode("ascii"))
        self.p.stdin.flush()

    def get_one(self, addr):
        self.command_counter += 1
        self.check_for_process_restart()
        cmd = "0x{:08x} w\n\n".format(addr)
        self.p.stdin.write(cmd.encode("ascii"))
        self.p.stdin.flush()
        ret = self.p.stdout.readline().decode("ascii")
        return int(ret.split('[10]')[-1].strip())


class PitayaTB(PitayaCSR):
    def __init__(self, x=None):
        from transfer import Filter, CsrThread
        from gateware.chains import FastChain

        p = FastChain(14, 25, 18)
        p.x = p.adc
        p.y = p.dac

        if x is None:
            x = np.random.uniform(-.8, .8, 1<<12)
        self.io = Filter(p, x)
        self.csr = CsrThread(self.io, p.get_csrs())
        self.offset = 0x40300000

    def start(self):
        self.csr.sim.start()

    def stop(self, n=0):
        self.csr.queue.append(n)
        self.csr.queue.append(None)
        self.csr.sim.join()

    def set_one(self, addr, value):
        addr -= self.offset
        return self.csr.write(addr//4, value)

    def get_one(self, addr):
        addr -= self.offset
        return self.csr.read(addr//4)


if __name__ == "__main__":
    p = PitayaSSH()

    new = dict(
        fast_a_x_tap=2,
        fast_a_demod_delay=0xc00,
        fast_a_x_clear_en=0, #p.states("fast_a_x_sat"),
        fast_a_brk=1,
        fast_a_dx_sel=p.signal("scopegen_dac_a"),#p.signal("zero"),
        fast_a_y_tap=1,
        fast_a_rx_sel=p.signal('zero'),#p.signal("fast_b_x"),
        fast_a_y_hold_en=p.states("fast_a_unlocked"),
        fast_a_y_clear_en=p.states("fast_a_y_railed"),
        fast_a_relock_run=0,
        fast_a_relock_en=p.states(),
        fast_a_relock_step=200,
        fast_a_relock_min=4000,
        fast_a_relock_max=8191,
        fast_a_sweep_run=0,
        #fast_a_sweep_step=1000,
        fast_a_sweep_step=0x00ffff,
        fast_a_sweep_min=-8192/2,
        fast_a_sweep_max=8191/2,

        fast_a_mod_amp=0x0e00,
        fast_a_mod_freq=0x10000000,
        #fast_a_mod_amp=0x0,
        # 0x10000000 ~= 8 MHz
        #fast_a_mod_freq=0x00000000,
        fast_a_dy_sel=p.signal('zero'),#p.signal("scopegen_dac_a"),
        fast_a_y_limit_min=-8192,
        fast_a_y_limit_max=8191,
        # 50uV rms / sqrt(Hz), 550mV rms/sqrt(125MHz)
        #
        fast_b_x_tap=1,
        fast_b_brk=1,
        fast_b_dx_sel=p.signal("zero"),
        fast_b_y_tap=1,
        fast_b_y_clear_en=p.states("fast_b_y_railed"),
        #fast_b_mod_amp=0x0000,
        #fast_b_mod_freq=0x00001234,
        fast_b_mod_amp=0x0400,
        fast_b_mod_freq=0x10000000,
        fast_b_dy_sel=p.signal("zero"),

        fast_b_sweep_run=0,
        #fast_a_sweep_step=1000,
        fast_b_sweep_step=0x00ffff,
        fast_b_sweep_min=-8192/2,
        fast_b_sweep_max=8191/2,

        slow_a_brk=1,
        slow_a_dx_sel=p.signal("scopegen_dac_a"),
        slow_a_clear_en=p.states("slow_a_sat"),
        slow_a_y_limit_min=0,

        noise_bits=25,
        #scopegen_adc_a_sel=p.signal("fast_a_x"),
        scopegen_adc_a_sel=p.signal("fast_a_x"),
        scopegen_adc_b_sel=p.signal("fast_a_y"),

        gpio_p_oes=0,
        gpio_n_oes=0xff,
        gpio_n_do0_en=p.states("fast_a_x_sat"),
        gpio_n_do1_en=p.states("fast_a_y_sat"),
        gpio_n_do2_en=p.states("fast_a_y_railed"),
        gpio_n_do3_en=p.states("fast_a_unlocked"),
        gpio_n_do4_en=p.states("slow_a_sat"),
        gpio_n_do5_en=p.states("slow_a_railed"),
    )

    for k, v in sorted(new.items()):
        p.set(k, int(v))

    b, a = make_filter("P", k=1)
    p.set_iir("fast_a_iir_a", b, a)
    p.set_iir("fast_a_iir_b", b, a)
    n = "fast_a_iir_c"
    p.set_iir("fast_a_iir_c", *make_filter("PI", k=.91, f=5e-5))
    p.set_iir("fast_a_iir_d", b, a)
    p.set_iir("fast_a_iir_e", *make_filter("PI", k=-.01, f=1e-5))
    p.set_iir("fast_b_iir_a", *make_filter("LP", k=1, f=1e-2))
    p.set_iir("fast_b_iir_c", *make_filter("LP", k=100, f=5e-6))
    p.set_iir("fast_b_iir_d", *make_filter("LP", k=1, f=1e-5))
    p.set_iir("fast_b_iir_e", *make_filter("LP", k=80, f=2e-8, q=40.))
    p.set_iir("slow_a_iir", *make_filter("PI", k=-.01, f=1e-4), z=0)



import os
import sys
import pickle
import _thread
import threading
from rpyc import Service
from time import sleep, time
from random import random
from rpyc.utils.server import OneShotServer
from PyRedPitaya.board import RedPitaya

sys.path += ['../../']
from csr import make_filter, PitayaLocal
from linien.config import ACQUISITION_PORT


def shutdown():
    _thread.interrupt_main()
    os._exit(0)


class DataAcquisitionService(Service):
    def __init__(self):
        self.r = RedPitaya()
        self.csr = PitayaLocal()
        self.csr_queue = []
        self.csr_iir_queue = []

        self.data = pickle.dumps(None)
        self.data_hash = None

        super(DataAcquisitionService, self).__init__()

        self.exposed_set_ramp_speed(9)
        self.locked = False

        self.run()

    def run(self, trigger_delay=16384):
        def run_acquiry_loop():
            while True:
                while self.csr_queue:
                    key, value = self.csr_queue.pop(0)
                    self.csr.set(key, value)

                while self.csr_iir_queue:
                    args = self.csr_iir_queue.pop(0)
                    self.csr.set_iir(*args)

                # copied from https://github.com/RedPitaya/RedPitaya/blob/14cca62dd58f29826ee89f4b28901602f5cdb1d8/api/src/oscilloscope.c#L115
                # check whether scope was triggered
                if (self.r.scope.read(0x1<<2) & 0x4) > 0 and not self.locked:
                    sleep(.05)
                    continue

                data = [
                    [int(i) for i in channel[:]]
                    for channel in
                    (self.r.scope.data_ch1, self.r.scope.data_ch2)
                ]

                slow_out = self.csr.get('root_slow_value')
                slow_out = slow_out if slow_out <= 8191 else slow_out - 16384
                data += [slow_out]

                self.r.scope.rearm(trigger_source=6)
                self.r.scope.data_decimation = 2 ** self.ramp_speed
                self.r.scope.trigger_delay = trigger_delay - 1

                self.data = pickle.dumps(data)
                self.data_hash = random()

        self.t = threading.Thread(target=run_acquiry_loop, args=())
        self.t.daemon = True
        self.t.start()

    def exposed_get_data_hash(self):
        return self.data_hash

    def exposed_return_data(self):
        return self.data[:]

    def exposed_set_ramp_speed(self, speed):
        self.ramp_speed = speed

    def exposed_set_lock_status(self, locked):
        self.locked = locked

    def exposed_set_csr(self, key, value):
        self.csr_queue.append((key, value))

    def exposed_set_iir_csr(self, *args):
        self.csr_iir_queue.append(args)



if __name__ == '__main__':
    t = OneShotServer(DataAcquisitionService(), port=ACQUISITION_PORT)
    t.start()


import numpy as np
from time import sleep, time

from csr import PitayaCSR, make_filter
from utils import start_nginx, stop_nginx, twos_complement
from linien.config import DEFAULT_RAMP_SPEED
from linien.common import convert_channel_mixing_value, \
    LOW_PASS_FILTER, HIGH_PASS_FILTER, ANALOG_OUT0
from linien.server.acquisition import AcquisitionMaster


class Registers:
    """This class provides low-level access to the FPGA registers."""
    def __init__(self, host=None, user=None, password=None):
        self.host = host
        self.user = user
        self.password = password
        self.acquisition = None

    def connect(self, control, parameters):
        """Starts a process that can be used to control FPGA registers."""
        self.control = control
        self.parameters = parameters

        self.csr = PitayaCSR()

        def lock_status_changed(v):
            if self.acquisition is not None:
                self.acquisition.lock_status_changed(v)

        self.parameters.lock.change(lock_status_changed)
        use_ssh = self.host is not None and self.host not in ('localhost', '127.0.0.1')
        self.acquisition = AcquisitionMaster(use_ssh, self.host)

    def run_data_acquisition(self, on_change):
        """Starts a background process that continuously reads out error /
        control signal of the FPGA. For every result, `on_change` is called."""
        self.acquisition.run_data_acquisition(on_change)

    def write_registers(self):
        """Writes data from `parameters` to the FPGA."""
        params = dict(self.parameters)

        _max = lambda val: val if np.abs(val) <= 8191 else (8191 * val / np.abs(val))
        phase_to_delay = lambda phase: int(phase / 360 * (1<<14))

        if not params['dual_channel']:
            factor_a = 256
            factor_b = 0
        else:
            value = params['channel_mixing']
            factor_a, factor_b = convert_channel_mixing_value(value)

        lock = params['lock']
        lock_changed = lock != self.control.exposed_is_locked
        self.control.exposed_is_locked = lock

        new = dict(
            root_sweep_run=0 if lock else 1,
            root_sweep_step=int(
                DEFAULT_RAMP_SPEED * params['ramp_amplitude']
                / (2 ** params['ramp_speed'])
            ),
            root_sweep_min=-1 * _max(params['ramp_amplitude'] * 8191),
            root_sweep_max=_max(params['ramp_amplitude'] * 8191),

            root_mod_freq=params['modulation_frequency'],
            root_mod_amp=params['modulation_amplitude'],

            root_chain_a_factor=factor_a,
            root_chain_b_factor=factor_b,
            root_chain_a_offset=twos_complement(int(params['offset_a']), 14),
            root_chain_b_offset=twos_complement(int(params['offset_b']), 14),
            root_out_offset=int(params['center'] * 8191),
            root_combined_offset=twos_complement(params['combined_offset'], 14),

            root_control_channel=params['control_channel'],
            root_mod_channel=params['mod_channel'],
            root_sweep_channel=params['sweep_channel'],

            slow_pid_reset=not params['pid_on_slow_enabled'],

            # channel A
            fast_a_x_tap=2,
            fast_a_demod_delay=phase_to_delay(params['demodulation_phase_a']),
            fast_a_demod_multiplier=params['demodulation_multiplier_a'],
            fast_a_brk=0,
            fast_a_dx_sel=self.csr.signal('zero'),
            fast_a_y_tap=2,
            fast_a_dy_sel=self.csr.signal('zero'),

            fast_a_y_hold_en=self.csr.states(),
            fast_a_y_clear_en=self.csr.states(),
            fast_a_rx_sel=self.csr.signal('zero'),

            # channel B
            fast_b_x_tap=2,
            fast_b_demod_delay=phase_to_delay(params['demodulation_phase_b']),
            fast_b_demod_multiplier=params['demodulation_multiplier_b'],
            fast_b_brk=0,
            fast_b_dx_sel=self.csr.signal('zero'),
            fast_b_y_tap=1,
            fast_b_dy_sel=self.csr.signal('zero'),

            fast_b_y_hold_en=self.csr.states(),
            fast_b_y_clear_en=self.csr.states(),
            fast_b_rx_sel=self.csr.signal('zero'),

            # trigger on ramp
            scopegen_external_trigger=1,

            gpio_p_oes=0,
            gpio_n_oes=0b1,

            gpio_p_outs=0,
            gpio_n_outs=0b1 if params['lock'] else 0b0,

            gpio_n_do0_en=self.csr.signal('zero'),
            gpio_n_do1_en=self.csr.signal('zero'),

            root_slow_decimation=16
        )

        if lock:
            # display combined error signal and control signal
            new.update({
                'scopegen_adc_a_sel': self.csr.signal('root_combined_error_signal'),
                'scopegen_adc_b_sel': self.csr.signal('root_control_signal')
            })
        else:
            # display both demodulated error signals
            new.update({
                'scopegen_adc_a_sel': self.csr.signal("fast_a_y"),
                'scopegen_adc_b_sel': self.csr.signal("fast_b_y")
            })


        # filter out values that did not change
        new = dict(
            (k, v)
            for k, v in new.items()
            if (
                (k not in self.control._cached_data)
                or (self.control._cached_data.get(k) != v)
            )
        )
        self.control._cached_data.update(new)

        # pass ramp speed changes to acquisition process
        if 'root_sweep_step' in new:
            self.acquisition.set_ramp_speed(params['ramp_speed'])

        for k, v in new.items():
            self.set(k, int(v))

        if 'root_sweep_step' in new:
            # reset sweep for a short time if the scan range was changed
            # this is needed because otherwise it may take too long before
            # the new scan range is reached --> no scope trigger is sent
            self.set('root_sweep_run', 0)
            self.set('root_sweep_run', 1)

        kp = params['p']
        ki = params['i']
        kd = params['d']
        slope = params['target_slope_rising']
        control_channel, sweep_channel = params['control_channel'], params['sweep_channel']

        def channel_polarity(channel):
            return (
                params['polarity_fast_out1'], params['polarity_fast_out2'],
                params['polarity_analog_out0']
            )[channel]

        if control_channel != sweep_channel:
            if channel_polarity(control_channel) != channel_polarity(sweep_channel):
                slope = not slope

        slow_strength = params['pid_on_slow_strength'] \
            if params['pid_on_slow_enabled'] \
            else 0
        slow_slope = slope
        if channel_polarity(ANALOG_OUT0) != channel_polarity(control_channel):
            slow_slope = not slow_slope

        for chain in ('a', 'b'):
            for iir_idx in range(2):
                iir_name = 'fast_%s_iir_%s' % (chain, ('c', 'd')[iir_idx])

                filter_enabled = params['filter_%d_enabled_%s' % (iir_idx + 1, chain)]
                filter_type = params['filter_%d_type_%s' % (iir_idx + 1, chain)]
                filter_frequency = params['filter_%d_frequency_%s' % (iir_idx + 1, chain)]
                base_freq = 125e6

                if not filter_enabled:
                    self.set_iir(iir_name, *make_filter('P', k=1))
                else:
                    if filter_type == LOW_PASS_FILTER:
                        self.set_iir(iir_name, *make_filter(
                            'LP', f=filter_frequency / base_freq, k=1
                        ))
                    elif filter_type == HIGH_PASS_FILTER:
                        self.set_iir(iir_name, *make_filter(
                            'HP', f=filter_frequency / base_freq, k=1
                        ))
                    else:
                        raise Exception('unknown filter %s for %s' (filter_type, iir_name))

        if lock_changed:
            if lock:
                # set PI parameters
                self.set_pid(kp, ki, kd, slope, reset=0)
                self.set_slow_pid(slow_strength, slow_slope, reset=0)
            else:
                self.set_pid(0, 0, 0, slope, reset=1)
                self.set_slow_pid(0, slow_slope, reset=1)

                self.set_iir("fast_a_iir_a", *make_filter('P', k=1))
                self.set_iir("fast_b_iir_a", *make_filter('P', k=1))

        else:
            if lock:
                # set new PI parameters
                self.set_pid(kp, ki, kd, slope)
                self.set_slow_pid(slow_strength, slow_slope)

    def set_pid(self, p, i, d, slope, reset=None):
        sign = -1 if slope else 1
        self.set('root_pid_kp', p * sign)
        self.set('root_pid_ki', i * sign)
        self.set('root_pid_kd', d * sign)

        if reset is not None:
            self.set('root_pid_reset', reset)

    def set_slow_pid(self, strength, slope, reset=None):
        sign = -1 if slope else 1
        self.set('slow_pid_ki', strength * sign)

        if reset is not None:
            self.set('slow_pid_reset', reset)

    def set(self, key, value):
        self.acquisition.set_csr(key, value)

    def set_iir(self, *args):
        self.acquisition.set_iir_csr(*args)


import os
import pickle
import appdirs


COLORS = {
    'spectroscopy1': (200, 0, 0, 200),
    'spectroscopy2': (0, 200, 0, 200),
    'spectroscopy_combined': (0, 0, 200, 200),
    'control_signal': (200, 0, 0, 200),
    'control_signal_history': (0, 200, 0, 200),
    'slow_history': (200, 200, 0, 200)
}


def get_data_folder():
    folder_name = appdirs.user_data_dir('linien')

    if not os.path.exists(folder_name):
        os.makedirs(folder_name)

    return folder_name


def get_devices_filename():
    return os.path.join(
        get_data_folder(),
        'devices'
    )


def save_device_data(devices):
    with open(get_devices_filename(), 'wb') as f:
        pickle.dump(devices, f)


def load_device_data():
    try:
        with open(get_devices_filename(), 'rb') as f:
            devices = pickle.load(f)
    except FileNotFoundError:
        devices = []

    return devices


def save_parameter(device_key, param, value, delete=False):
    devices = load_device_data()
    device = [d for d in devices if d['key'] == device_key][0]
    device.setdefault('params', {})

    if not delete:
        device['params'][param] = value
    else:
        try:
            del device['params'][param]
        except KeyError:
            pass

    save_device_data(devices)


def get_saved_parameters(device_key):
    devices = load_device_data()
    device = [d for d in devices if d['key'] == device_key][0]
    device.setdefault('params', {})
    return device['params']


import uuid
from PyQt5 import QtGui, QtWidgets
from linien.client.config import load_device_data, save_device_data
from linien.client.widgets import CustomWidget


class NewDeviceDialog(QtWidgets.QDialog, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_ui('new_device_dialog.ui')

    def add_new_device(self):
        device = {
            'key':  uuid.uuid4().hex,
            'name': self.ids.deviceName.text(),
            'host': self.ids.host.text(),
            'username': self.ids.username.text(),
            'password': self.ids.password.text(),
            'params': {}
        }
        devices = load_device_data() + [device]
        save_device_data(devices)


import linien

from time import sleep
from socket import gaierror
from plumbum import colors
from traceback import print_exc
from PyQt5.QtCore import QThread, pyqtSignal
from paramiko.ssh_exception import AuthenticationException

from linien.config import SERVER_PORT
from linien.client.utils import run_server
from linien.client.config import save_parameter, get_saved_parameters
from linien.client.exceptions import GeneralConnectionErrorException, \
    InvalidServerVersionException, ServerNotInstalledException
from linien.communication.client import BaseClient


MHz = 0x10000000 / 8
Vpp = ((1<<14) - 1) / 4


class ConnectionThread(QThread):
    connected = pyqtSignal(object)
    server_not_installed = pyqtSignal()
    invalid_server_version = pyqtSignal(str, str)
    authentication_exception = pyqtSignal()
    general_connection_error = pyqtSignal()
    exception = pyqtSignal()
    connection_lost = pyqtSignal()

    def __init__(self, device):
        super().__init__()

        self.device = device

    def run(self):
        try:
            conn = Connection(self.device, self.on_connection_lost)
            self.connected.emit(conn)

        except ServerNotInstalledException:
            return self.server_not_installed.emit()

        except InvalidServerVersionException as e:
            return self.invalid_server_version.emit(e.remote_version, e.client_version)

        except AuthenticationException:
            return self.authentication_exception.emit()

        except GeneralConnectionErrorException:
            return self.general_connection_error.emit()

        except Exception as e:
            print_exc()
            return self.exception.emit()

    def on_connection_lost(self):
        self.connection_lost.emit()


class Connection(BaseClient):
    def __init__(self, device, on_connection_lost):
        self.device = device
        self.host = device['host']
        self.user = device.get('username')
        self.password = device.get('password')

        if self.host in ('localhost', '127.0.0.1'):
            # RP is configured such that "localhost" doesn't point to
            # 127.0.0.1 in all cases
            self.host = '127.0.0.1'
        else:
            assert self.user and self.password

        super().__init__(self.host, SERVER_PORT, True, call_on_error=on_connection_lost)

    def connect(self, host, port, use_parameter_cache, call_on_error=None):
        self.connection = None

        i = -1
        server_was_started = False

        while True:
            i += 1

            try:
                print('try to connect to %s:%s' % (host, port))
                self._connect(host, port, use_parameter_cache, call_on_error=call_on_error)
                self.control = self.connection.root
                break
            except gaierror:
                # host not found
                print(colors.red | 'Error: host %s not found' % host)
                break
            except Exception as e:
                if i == 0:
                    print('server is not running. Launching it!')
                    server_was_started = True
                    run_server(host, self.user, self.password)
                    sleep(3)
                else:
                    if i < 20:
                        print('server still not running, waiting (this may take some time)...')
                        sleep(1)
                    else:
                        print_exc()
                        print(colors.red | 'Error: connection to the server could not be established')
                        break

        if self.connection is None:
            raise GeneralConnectionErrorException()

        # now check that the remote version is the same as ours
        remote_version = self.connection.root.exposed_get_server_version()
        client_version = linien.__version__

        if remote_version != client_version:
            raise InvalidServerVersionException(client_version, remote_version)

        print(colors.green | 'connected established!')

        if server_was_started:
            # without this sleep, parameter restoring sometimes crashed the sever
            sleep(1)
            self.restore_parameters()
        self.prepare_parameter_restoring()

    def disconnect(self):
        self.connection.close()

    def prepare_parameter_restoring(self):
        """Listens for changes of some parameters and permanently saves their
        values on the client's disk. This data can be used to restore the status
        later, if the client tries to connect to the server but it doesn't run
        anymore."""
        params = self.parameters.remote.exposed_get_restorable_parameters()

        for param in params:
            def on_change(value, param=param):
                save_parameter(self.device['key'], param, value)

            getattr(self.parameters, param).change(on_change)

    def restore_parameters(self):
        device_key = self.device['key']
        params = get_saved_parameters(device_key)
        print('restoring parameters')

        for k, v in params.items():
            if hasattr(self.parameters, k):
                getattr(self.parameters, k).value = v
            else:
                # this may happen if the settings were written with a different
                # version of linien.
                print('unable to restore parameter %s. Delete the cached value.' % k)
                save_parameter(device_key, k, None, delete=True)

        self.control.write_data()


import paramiko
from plumbum import colors

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QApplication, QWidget, QListWidget, QVBoxLayout, QLabel, QPushButton, QListWidgetItem, \
    QHBoxLayout, QDialog, QMessageBox
from pyqtgraph import QtCore

from linien.client.utils import connect_ssh


class SSHCommandOutputWidget(QListWidget):
    command_ended = pyqtSignal()

    def __init__(self, *args):
        super().__init__(*args)
        self.setSelectionMode(self.NoSelection)

    def execute(self, ssh, command):
        self.stdin, self.stdout, self.stderr = ssh.exec_command(command, bufsize=0, get_pty=True)
        self.addItem('>>> %s' % command)
        self.show_output()

    def show_output(self):
        if self.stdout.channel.exit_status_ready():
            return self.command_ended.emit()
        else:
            for output in (self.stdout, self.stderr):
                toread = len(output.channel.in_buffer)
                if toread == 0:
                    continue
                buf = output.read(toread).decode('utf8').rstrip('\n')

                for part in buf.split('\n'):
                    for subpart in part.split('\r'):
                        subpart = subpart.strip('\n').strip('\r').strip('\r\n')
                        if subpart:
                            print(
                                (colors.red if output == self.stderr else colors.reset)
                                | subpart
                            )
                            self.addItem(subpart)
                self.scrollToBottom()

        QtCore.QTimer.singleShot(1000, self.show_output)


def execute_command(parent, host, user, password, command, callback):
    print((colors.bold | 'Execute command: ') + command)

    ssh = connect_ssh(host, user, password)

    window = QDialog(parent)
    window.resize(800, 600)
    window_layout = QVBoxLayout(window)

    widget = SSHCommandOutputWidget(parent)
    window_layout.addWidget(widget)
    window.setLayout(window_layout)
    window.setModal(True)
    window.setWindowModality(QtCore.Qt.WindowModal)
    window.show()

    def after_command():
        window.hide()
        callback()
    widget.command_ended.connect(after_command)

    widget.execute(ssh, command)

    return window


class LoadingDialog(QMessageBox):
    aborted = pyqtSignal()

    def __init__(self, parent, host):
        super().__init__(parent)

        self.setIcon(QMessageBox.Information)
        self.setText('Connecting to %s' % host)
        self.setWindowTitle('Connecting')
        self.setModal(True)
        self.setWindowModality(QtCore.Qt.WindowModal)
        self.setStandardButtons(QMessageBox.NoButton)
        self.show()

    def closeEvent(self, *args):
        self.aborted.emit()

    def keyPressEvent(self, event):
        key = event.key()
        if key == QtCore.Qt.Key_Escape:
            self.close()


def error_dialog(parent, error):
    return QMessageBox.question(
        parent, 'Error', error,
        QMessageBox.Ok,
        QMessageBox.Ok
    )


def question_dialog(parent, question):
    reply = QMessageBox.question(
        parent, 'Error', question,
        QMessageBox.Yes | QMessageBox.No,
        QMessageBox.Yes
    )

    return reply == QMessageBox.Yes


import pickle
import traceback
import numpy as np
from time import sleep, time
from linien.common import determine_shift_by_correlation, get_lock_point, \
    control_signal_has_correct_amplitude, combine_error_signal, \
    check_plot_data, ANALOG_OUT0


class Autolock:
    """Spectroscopy autolock based on correlation."""
    def __init__(self, control, parameters):
        self.control = control
        self.parameters = parameters

        self.first_error_signal = None
        self.parameters.autolock_running.value = False
        self.should_watch_lock = False

        self.reset_properties()

    def reset_properties(self):
        self.history = []
        self.zoom_factor = 1
        self.N_at_this_zoom = 0
        self.last_shifts_at_this_zoom = None
        self.skipped = 0
        self.parameters.autolock_failed.value = False
        self.parameters.autolock_locked.value = False
        self.parameters.autolock_watching.value = False

    def run(self, x0, x1, spectrum, should_watch_lock=False, auto_offset=True):
        """Starts the autolock.

        If `should_watch_lock` is specified, the autolock continuously monitors
        the control and error signals after the lock was successful and tries to
        relock automatically using the spectrum that was recorded in the first
        run of the lock.
        """
        self.parameters.autolock_running.value = True
        self.x0, self.x1 = int(x0), int(x1)
        self.should_watch_lock = should_watch_lock
        self.auto_offset = auto_offset

        self.parameters.autolock_approaching.value = True
        self.record_first_error_signal(spectrum)

        self.initial_ramp_amplitude = self.parameters.ramp_amplitude.value
        self.parameters.autolock_initial_ramp_amplitude.value = self.initial_ramp_amplitude
        self.initial_ramp_center = self.parameters.center.value

        self.add_data_listener()

    def add_data_listener(self):
        self.parameters.to_plot.change(self.react_to_new_spectrum)

    def remove_data_listener(self):
        self.parameters.to_plot.remove_listener(self.react_to_new_spectrum)

    def react_to_new_spectrum(self, plot_data):
        """React to new spectrum data.

        If this is executed for the first time, a reference spectrum is
        recorded.

        If the autolock is approaching the desired line, a correlation
        function of the spectrum with the reference spectrum is calculated
        and the laser current is adapted such that the targeted line is centered.

        After this procedure is done, the real lock is turned on and after some
        time the lock is verified.

        If automatic relocking is desired, the control and error signals are
        continuously monitored after locking.
        """
        if self.parameters.pause_acquisition.value:
            return

        if plot_data is None or not self.parameters.autolock_running.value:
            return

        plot_data = pickle.loads(plot_data)
        if plot_data is None:
            return

        is_locked = self.parameters.lock.value

        if not check_plot_data(is_locked, plot_data):
            return

        if is_locked:
            error_signal = plot_data['error_signal']
            control_signal = plot_data['control_signal']
        else:
            combined_error_signal = combine_error_signal(
                (plot_data['error_signal_1'], plot_data['error_signal_2']),
                self.parameters.dual_channel.value,
                self.parameters.channel_mixing.value
            )

        try:
            if self.parameters.autolock_approaching.value:
                # we have already recorded a spectrum and are now approaching
                # the line by decreasing the scan range and adapting the
                # center current multiple times.

                if self.skipped < 1:
                    # after every step, we skip some data in order to let
                    # the laser equilibrate
                    self.skipped += 1
                    return

                self.skipped = 0

                return self.approach_line(combined_error_signal)

            elif self.parameters.autolock_watching.value:
                # the laser was locked successfully before. Now we check
                # periodically whether the laser is still in lock

                return self.watch_lock(error_signal, control_signal)

            else:
                # we are done with approaching and have started the lock.
                # skip some data and check whether we really are in lock
                # afterwards.

                if self.skipped < 3:
                    self.skipped += 1
                    return

                self.skipped = 0

                return self.after_lock(control_signal, plot_data.get('slow'))

        except Exception:
            traceback.print_exc()
            self.exposed_stop()

    def record_first_error_signal(self, error_signal):
        mean_signal, target_slope_rising, target_zoom, rolled_error_signal = \
            get_lock_point(error_signal, self.x0, self.x1)

        if self.auto_offset:
            self.parameters.combined_offset.value = -1 * mean_signal
            rolled_error_signal -= int(mean_signal)

        self.parameters.target_slope_rising.value = target_slope_rising
        self.control.exposed_write_data()

        self.target_zoom = target_zoom
        self.first_error_signal = rolled_error_signal

    def approach_line(self, error_signal):
        shift, zoomed_ref, zoomed_err = determine_shift_by_correlation(
            self.zoom_factor, self.first_error_signal, error_signal
        )
        shift *= self.initial_ramp_amplitude
        self.history.append((zoomed_ref, zoomed_err))

        self.control.exposed_write_data()
        self.history.append('shift %f' % (-1 * shift))

        self.parameters.center.value -= shift
        self.control.exposed_write_data()

        if self.N_at_this_zoom > 50:
            raise Exception('max number of N_at_this_zoom exceeded: %d' % self.N_at_this_zoom)

        if self.last_shifts_at_this_zoom is not None:
            # check that the drift is slow
            # this is needed for systems that only react slowly to changes in
            # input parameters. In this case, we have to wait until the reaction
            # to the last input is done.
            drift_is_slow = (np.abs(shift - self.last_shifts_at_this_zoom[-1])) \
                            < (0.05 * self.parameters.ramp_amplitude.value)

            # if the drift changed sign, it is no drift but noise... This means
            # that we should go on
            if len(self.last_shifts_at_this_zoom) < 2:
                drift_changed_sign = False
            else:
                drift_changed_sign = np.sign(shift - self.last_shifts_at_this_zoom[-1]) \
                    != np.sign(self.last_shifts_at_this_zoom[-1] - self.last_shifts_at_this_zoom[-2])

            if drift_is_slow or drift_changed_sign:
                self.N_at_this_zoom = 0
                self.last_shifts_at_this_zoom = None

                zoom_step = 2
                self.zoom_factor *= zoom_step

                self.control.pause_acquisition()

                self.parameters.ramp_amplitude.value /= zoom_step
                self.control.exposed_write_data()

                if self.zoom_factor >= self.target_zoom:
                    self.parameters.autolock_approaching.value = False
                    self.control.exposed_start_lock()

                self.control.continue_acquisition()

        self.N_at_this_zoom += 1
        self.last_shifts_at_this_zoom = self.last_shifts_at_this_zoom or []
        self.last_shifts_at_this_zoom.append(shift)

    def after_lock(self, control_signal, slow_out):
        """After locking, this method checks whether the laser really is locked.

        If desired, it automatically tries to relock if locking failed, or
        starts a watcher that does so over and over again.
        """
        def check_whether_in_lock(control_signal):
            """
            The laser is considered in lock if the mean value of the control
            signal is within the boundaries of the smalles current ramp we had
            before turning on the lock.
            """
            center = self.parameters.center.value
            ampl = self.parameters.ramp_amplitude.value

            slow_ramp = self.parameters.sweep_channel.value == ANALOG_OUT0
            slow_pid = self.parameters.pid_on_slow_enabled.value

            if not slow_ramp and not slow_pid:
                mean = np.mean(control_signal) / 8192

                return (center - ampl) <= mean <= (center + ampl)
            else:
                if slow_pid and not slow_ramp:
                    # we cannot handle this case. Just assume the laser is locked.
                    return True

                return (center - ampl) <= slow_out / 8192 <= (center + ampl)

        self.parameters.autolock_locked.value = check_whether_in_lock(control_signal)

        if self.parameters.autolock_locked.value and self.should_watch_lock:
            # we start watching the lock status from now on.
            # this is done in `react_to_new_spectrum()` which is called regularly.
            self.parameters.autolock_watching.value = True
        else:
            self.remove_data_listener()

            if not self.parameters.autolock_locked.value:
                if self.should_watch_lock:
                    return self.relock()

                self.control.exposed_reset_scan()
                self.parameters.autolock_failed.value = True
                self.remove_data_listener()

            self.parameters.autolock_running.value = False

    def watch_lock(self, error_signal, control_signal):
        """Check whether the laser is still in lock and init a relock if not."""
        mean = np.abs(np.mean(control_signal) / 8192)
        still_in_lock = mean < 0.9

        if not still_in_lock:
            self.relock()

    def relock(self):
        """
        Relock the laser using the reference spectrum recorded in the first
        locking approach.
        """
        self.reset_properties()
        self.control.pause_acquisition()

        self.parameters.autolock_running.value = True
        self.parameters.autolock_approaching.value = True

        self.parameters.center.value = self.initial_ramp_center
        self.parameters.ramp_amplitude.value = self.initial_ramp_amplitude
        self.control.exposed_start_ramp()

        self.control.continue_acquisition()

        # add a listener that listens for new spectrum data and consequently
        # tries to relock.
        self.add_data_listener()

    def exposed_stop(self):
        """Abort any operation."""
        self.parameters.autolock_failed.value = True
        self.parameters.autolock_running.value = False
        self.parameters.autolock_locked.value = False
        self.parameters.autolock_approaching.value = False
        self.parameters.autolock_watching.value = False
        self.remove_data_listener()

        self.control.exposed_reset_scan()


"""
    communication.client
    ~~~~~~~~~~~~~~~~~~~~

    Contains the client that can be used to access a service with easy access
    to the server's parameters.
"""
import rpyc
import uuid
from plumbum import colors
from .remote_parameters import RemoteParameters


class ClientService(rpyc.Service):
    def __init__(self, uuid):
        super().__init__()

        self.exposed_uuid = uuid


class BaseClient:
    def __init__(self, server, port, use_parameter_cache, call_on_error=None):
        self.use_parameter_cache = use_parameter_cache
        self.uuid = uuid.uuid4().hex

        self.client_service = ClientService(self.uuid)

        self.connect(server, port, use_parameter_cache, call_on_error=call_on_error)
        self.connected = True

    def connect(self, server, port, use_parameter_cache, call_on_error=None):
        return self._connect(server, port, use_parameter_cache, call_on_error=call_on_error)

    def _connect(self, server, port, use_parameter_cache, call_on_error=None):
        self.connection = rpyc.connect(server, port, service=self.client_service)

        cls = RemoteParameters
        if call_on_error:
            cls = self.catch_network_errors(cls, call_on_error)

        self.parameters = cls(
            self.connection.root,
            self.uuid,
            use_parameter_cache
        )

    def catch_network_errors(self, cls, call_on_error):
        function_type = type(lambda x: x)
        for attr_name in dir(cls):
            if not attr_name.startswith('__'):
                attr = getattr(cls, attr_name)

                if isinstance(attr, function_type):
                    method = attr
                    def wrapped(*args, method=method, **kwargs):
                        try:
                            return method(*args, **kwargs)
                        except (EOFError,):
                            print(colors.red | 'Connection lost')
                            self.stop()
                            call_on_error()
                            raise

                    setattr(cls, attr_name, wrapped)

        return cls

    def stop(self):
        self.connected = False


# Copyright 2014-2015 Robert Jordens <jordens@gmail.com>
#
# This file is part of redpid.
#
# redpid is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# redpid is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with redpid.  If not, see <http://www.gnu.org/licenses/>.

from math import log2, ceil, pi
import warnings

import scipy.signal


def make_filter(name, k=1., f=0., g=1e20, q=.5):
    f *= pi

    if name == "LP": # k/(s + 1)
        b = [f*k/(f + 2), f*k/(f + 2)]
        a = [1, (f - 2)/(f + 2)]

    elif name == "HP": # k/(1 + 1/s)
        b = [2*k/(f + 2), 2*k/(-f - 2)]
        a = [1, (f - 2)/(f + 2)]

    elif name == "AP": # k*(s - 1)/(s + 1)
        b = [k*(-f + 2)/(f + 2), -k]
        a = [1, (f - 2)/(f + 2)]

    elif name == "I": # k/s
        b = [f*k/2, f*k/2]
        a = [1, -1]

    elif name == "PI": # k*(s + 1)/(s + 1/g)
        b = [g*k*(f + 2)/(f + 2*g), g*k*(f - 2)/(f + 2*g)]
        a = [1, (f - 2*g)/(f + 2*g)]

    elif name == "P": # k
        b = [k]
        a = [1]

    elif name == "PD": # k*(s + 1)/(1 + s/g)
        b = [g*k*(f + 2)/(f*g + 2), g*k*(f - 2)/(f*g + 2)]
        a = [1, (f*g - 2)/(f*g + 2)]

    elif name == "LP2": # k/(s**2 + 1 + s/q)
        b = [f**2*k*q/(f**2*q + 2*f + 4*q), 2*f**2*k*q/(f**2*q + 2*f + 4*q),
                f**2*k*q/(f**2*q + 2*f + 4*q)]
        a = [1, 2*q*(f**2 - 4)/(f**2*q + 2*f + 4*q),
                (f**2*q - 2*f + 4*q)/(f**2*q + 2*f + 4*q)]

    elif name == "HP2": # k/(1 + s**(-2) + 1/(q*s))
        b = [4*k*q/(f**2*q + 2*f + 4*q), 8*k*q/(-f**2*q - 2*f - 4*q),
                4*k*q/(f**2*q + 2*f + 4*q)]
        a = [1, 2*q*(f**2 - 4)/(f**2*q + 2*f + 4*q),
                (f**2*q - 2*f + 4*q)/(f**2*q + 2*f + 4*q)]

    elif name == "NOTCH": # k*(s**2 + 1)/(s**2 + 1 + s/q)
        b = [k*q*(f**2 + 4)/(f**2*q + 2*f + 4*q),
                2*k*q*(f**2 - 4)/(f**2*q + 2*f + 4*q),
                k*q*(f**2 + 4)/(f**2*q + 2*f + 4*q)]
        a = [1, 2*q*(f**2 - 4)/(f**2*q + 2*f + 4*q),
                (f**2*q - 2*f + 4*q)/(f**2*q + 2*f + 4*q)]

    elif name == "IHO": # 4*k*(s + 1/s + 1/q)/((1 + s/g)*(f*g + 2))
        b = [2*g*k*(f**2*q + 2*f + 4*q)/(q*(f**2*g**2 + 4*f*g + 4)),
                4*g*k*(f**2 - 4)/(f**2*g**2 + 4*f*g + 4),
                2*g*k*(f**2*q - 2*f + 4*q)/(q*(f**2*g**2 + 4*f*g + 4))]
        a = [1, 4/(-f*g - 2), (-f*g + 2)/(f*g + 2)]

    return b, a


def quantize_filter(b, a, shift=None, width=25):
    b, a = [i/a[0] for i in b], [i/a[0] for i in a]

    if shift is None:
        shift = width
        for i in b + a:
            m = ceil(log2(abs(i)))
            if i > 0 and int(m) == m:
                m += 1
            shift = min(shift, int(width - 1 - m))
    s = 1 << shift
    
    b = [int(round(i*s)) for i in b]
    a = [int(round(i*s)) for i in a]

    m = 1<<(width - 1)
    for i in b + a:
        assert -m <= i < m, (hex(i), hex(m))

    z, p, k = scipy.signal.tf2zpk(b, a)
    if any(abs(_) > 1 for _ in p):
        warnings.warn("unstable filter: z={}, p={}, k={}".format(
            z, p, k), RuntimeWarning)

    return b, a, shift


def get_params(b, a, shift=None, width=25, interval=1):
    b, a, shift = quantize_filter(b, a, shift, width)
    params = {}
    for i, (ai, bi) in enumerate(zip(a, b)):
        params["a%i" % i] = int(-ai)
        params["b%i" % i] = int(bi)
    del params["a0"]
    #params["shift"] = shift
    return b, a, params



class InvalidServerVersionException(Exception):
    def __init__(self, client_version, remote_version):
        self.client_version = client_version
        self.remote_version = remote_version

        super().__init__()


class ServerNotInstalledException(Exception):
    pass


class GeneralConnectionErrorException(Exception):
    pass


csr_constants = {
    'fast_a_iir_a_shift': 23,
    'fast_a_iir_a_width': 25,
    'fast_a_iir_a_interval': 1,
    'fast_a_iir_a_latency': 2,
    'fast_a_iir_a_order': 1,
    'fast_a_iir_a_iterative': 0,
    'fast_a_iir_b_shift': 23,
    'fast_a_iir_b_width': 25,
    'fast_a_iir_b_interval': 5,
    'fast_a_iir_b_latency': 6,
    'fast_a_iir_b_order': 2,
    'fast_a_iir_b_iterative': 1,
    'fast_a_iir_c_shift': 23,
    'fast_a_iir_c_width': 25,
    'fast_a_iir_c_interval': 1,
    'fast_a_iir_c_latency': 2,
    'fast_a_iir_c_order': 1,
    'fast_a_iir_c_iterative': 0,
    'fast_a_iir_d_shift': 23,
    'fast_a_iir_d_width': 25,
    'fast_a_iir_d_interval': 1,
    'fast_a_iir_d_latency': 3,
    'fast_a_iir_d_order': 2,
    'fast_a_iir_d_iterative': 0,
    'fast_a_iir_e_shift': 23,
    'fast_a_iir_e_width': 25,
    'fast_a_iir_e_interval': 5,
    'fast_a_iir_e_latency': 6,
    'fast_a_iir_e_order': 2,
    'fast_a_iir_e_iterative': 1,
    'fast_b_iir_a_shift': 23,
    'fast_b_iir_a_width': 25,
    'fast_b_iir_a_interval': 1,
    'fast_b_iir_a_latency': 2,
    'fast_b_iir_a_order': 1,
    'fast_b_iir_a_iterative': 0,
    'fast_b_iir_b_shift': 23,
    'fast_b_iir_b_width': 25,
    'fast_b_iir_b_interval': 5,
    'fast_b_iir_b_latency': 6,
    'fast_b_iir_b_order': 2,
    'fast_b_iir_b_iterative': 1,
    'fast_b_iir_c_shift': 23,
    'fast_b_iir_c_width': 25,
    'fast_b_iir_c_interval': 1,
    'fast_b_iir_c_latency': 2,
    'fast_b_iir_c_order': 1,
    'fast_b_iir_c_iterative': 0,
    'fast_b_iir_d_shift': 23,
    'fast_b_iir_d_width': 25,
    'fast_b_iir_d_interval': 1,
    'fast_b_iir_d_latency': 3,
    'fast_b_iir_d_order': 2,
    'fast_b_iir_d_iterative': 0,
    'fast_b_iir_e_shift': 23,
    'fast_b_iir_e_width': 25,
    'fast_b_iir_e_interval': 5,
    'fast_b_iir_e_latency': 6,
    'fast_b_iir_e_order': 2,
    'fast_b_iir_e_iterative': 1,
    'root_sweep_shift': 18,
}

csr = {
    'dna_dna': (28, 0x000, 64, False),
    'fast_a_x_tap': (0, 0x000, 2, True),
    'fast_a_brk': (0, 0x001, 1, True),
    'fast_a_y_tap': (0, 0x002, 2, True),
    'fast_a_iir_a_z0': (0, 0x003, 27, True),
    'fast_a_iir_a_a1': (0, 0x007, 25, True),
    'fast_a_iir_a_b0': (0, 0x00b, 25, True),
    'fast_a_iir_a_b1': (0, 0x00f, 25, True),
    'fast_a_demod_delay': (0, 0x013, 32, True),
    'fast_a_demod_multiplier': (0, 0x017, 4, True),
    'fast_a_iir_b_z0': (0, 0x018, 52, True),
    'fast_a_iir_b_a1': (0, 0x01f, 25, True),
    'fast_a_iir_b_a2': (0, 0x023, 25, True),
    'fast_a_iir_b_b0': (0, 0x027, 25, True),
    'fast_a_iir_b_b1': (0, 0x02b, 25, True),
    'fast_a_iir_b_b2': (0, 0x02f, 25, True),
    'fast_a_x_limit_min': (0, 0x033, 25, True),
    'fast_a_x_limit_max': (0, 0x037, 25, True),
    'fast_a_iir_c_z0': (0, 0x03b, 27, True),
    'fast_a_iir_c_a1': (0, 0x03f, 25, True),
    'fast_a_iir_c_b0': (0, 0x043, 25, True),
    'fast_a_iir_c_b1': (0, 0x047, 25, True),
    'fast_a_iir_d_z0': (0, 0x04b, 27, True),
    'fast_a_iir_d_a1': (0, 0x04f, 25, True),
    'fast_a_iir_d_a2': (0, 0x053, 25, True),
    'fast_a_iir_d_b0': (0, 0x057, 25, True),
    'fast_a_iir_d_b1': (0, 0x05b, 25, True),
    'fast_a_iir_d_b2': (0, 0x05f, 25, True),
    'fast_a_iir_e_z0': (0, 0x063, 52, True),
    'fast_a_iir_e_a1': (0, 0x06a, 25, True),
    'fast_a_iir_e_a2': (0, 0x06e, 25, True),
    'fast_a_iir_e_b0': (0, 0x072, 25, True),
    'fast_a_iir_e_b1': (0, 0x076, 25, True),
    'fast_a_iir_e_b2': (0, 0x07a, 25, True),
    'fast_a_y_limit_min': (0, 0x07e, 14, True),
    'fast_a_y_limit_max': (0, 0x080, 14, True),
    'fast_a_x_clr': (0, 0x082, 1, True),
    'fast_a_x_max': (0, 0x083, 25, False),
    'fast_a_x_min': (0, 0x087, 25, False),
    'fast_a_y_clr': (0, 0x08b, 1, True),
    'fast_a_y_max': (0, 0x08c, 25, False),
    'fast_a_y_min': (0, 0x090, 25, False),
    'fast_a_x_hold_en': (0, 0x094, 17, True),
    'fast_a_x_clear_en': (0, 0x097, 17, True),
    'fast_a_y_hold_en': (0, 0x09a, 17, True),
    'fast_a_y_clear_en': (0, 0x09d, 17, True),
    'fast_a_dx_sel': (0, 0x0a0, 4, True),
    'fast_a_dy_sel': (0, 0x0a1, 4, True),
    'fast_a_rx_sel': (0, 0x0a2, 4, True),
    'fast_b_x_tap': (1, 0x000, 2, True),
    'fast_b_brk': (1, 0x001, 1, True),
    'fast_b_y_tap': (1, 0x002, 2, True),
    'fast_b_iir_a_z0': (1, 0x003, 27, True),
    'fast_b_iir_a_a1': (1, 0x007, 25, True),
    'fast_b_iir_a_b0': (1, 0x00b, 25, True),
    'fast_b_iir_a_b1': (1, 0x00f, 25, True),
    'fast_b_demod_delay': (1, 0x013, 32, True),
    'fast_b_demod_multiplier': (1, 0x017, 4, True),
    'fast_b_iir_b_z0': (1, 0x018, 52, True),
    'fast_b_iir_b_a1': (1, 0x01f, 25, True),
    'fast_b_iir_b_a2': (1, 0x023, 25, True),
    'fast_b_iir_b_b0': (1, 0x027, 25, True),
    'fast_b_iir_b_b1': (1, 0x02b, 25, True),
    'fast_b_iir_b_b2': (1, 0x02f, 25, True),
    'fast_b_x_limit_min': (1, 0x033, 25, True),
    'fast_b_x_limit_max': (1, 0x037, 25, True),
    'fast_b_iir_c_z0': (1, 0x03b, 27, True),
    'fast_b_iir_c_a1': (1, 0x03f, 25, True),
    'fast_b_iir_c_b0': (1, 0x043, 25, True),
    'fast_b_iir_c_b1': (1, 0x047, 25, True),
    'fast_b_iir_d_z0': (1, 0x04b, 27, True),
    'fast_b_iir_d_a1': (1, 0x04f, 25, True),
    'fast_b_iir_d_a2': (1, 0x053, 25, True),
    'fast_b_iir_d_b0': (1, 0x057, 25, True),
    'fast_b_iir_d_b1': (1, 0x05b, 25, True),
    'fast_b_iir_d_b2': (1, 0x05f, 25, True),
    'fast_b_iir_e_z0': (1, 0x063, 52, True),
    'fast_b_iir_e_a1': (1, 0x06a, 25, True),
    'fast_b_iir_e_a2': (1, 0x06e, 25, True),
    'fast_b_iir_e_b0': (1, 0x072, 25, True),
    'fast_b_iir_e_b1': (1, 0x076, 25, True),
    'fast_b_iir_e_b2': (1, 0x07a, 25, True),
    'fast_b_y_limit_min': (1, 0x07e, 14, True),
    'fast_b_y_limit_max': (1, 0x080, 14, True),
    'fast_b_x_clr': (1, 0x082, 1, True),
    'fast_b_x_max': (1, 0x083, 25, False),
    'fast_b_x_min': (1, 0x087, 25, False),
    'fast_b_y_clr': (1, 0x08b, 1, True),
    'fast_b_y_max': (1, 0x08c, 25, False),
    'fast_b_y_min': (1, 0x090, 25, False),
    'fast_b_x_hold_en': (1, 0x094, 17, True),
    'fast_b_x_clear_en': (1, 0x097, 17, True),
    'fast_b_y_hold_en': (1, 0x09a, 17, True),
    'fast_b_y_clear_en': (1, 0x09d, 17, True),
    'fast_b_dx_sel': (1, 0x0a0, 4, True),
    'fast_b_dy_sel': (1, 0x0a1, 4, True),
    'fast_b_rx_sel': (1, 0x0a2, 4, True),
    'gpio_n_ins': (30, 0x000, 8, False),
    'gpio_n_outs': (30, 0x001, 8, True),
    'gpio_n_oes': (30, 0x002, 8, True),
    'gpio_n_state': (30, 0x003, 17, False),
    'gpio_n_state_clr': (30, 0x006, 1, True),
    'gpio_n_do0_en': (30, 0x007, 17, True),
    'gpio_n_do1_en': (30, 0x00a, 17, True),
    'gpio_n_do2_en': (30, 0x00d, 17, True),
    'gpio_n_do3_en': (30, 0x010, 17, True),
    'gpio_n_do4_en': (30, 0x013, 17, True),
    'gpio_n_do5_en': (30, 0x016, 17, True),
    'gpio_n_do6_en': (30, 0x019, 17, True),
    'gpio_n_do7_en': (30, 0x01c, 17, True),
    'gpio_p_ins': (31, 0x000, 8, False),
    'gpio_p_outs': (31, 0x001, 8, True),
    'gpio_p_oes': (31, 0x002, 8, True),
    'root_chain_a_factor': (8, 0x000, 9, True),
    'root_chain_b_factor': (8, 0x002, 9, True),
    'root_chain_a_offset': (8, 0x004, 14, True),
    'root_chain_b_offset': (8, 0x006, 14, True),
    'root_combined_offset': (8, 0x008, 14, True),
    'root_out_offset': (8, 0x00a, 14, True),
    'root_mod_channel': (8, 0x00c, 1, True),
    'root_control_channel': (8, 0x00d, 1, True),
    'root_sweep_channel': (8, 0x00e, 2, True),
    'root_slow_value': (8, 0x00f, 14, False),
    'root_mod_amp': (8, 0x011, 14, True),
    'root_mod_freq': (8, 0x013, 32, True),
    'root_sweep_step': (8, 0x017, 24, True),
    'root_sweep_min': (8, 0x01a, 14, True),
    'root_sweep_max': (8, 0x01c, 14, True),
    'root_sweep_run': (8, 0x01e, 1, True),
    'root_limit_error_signal_min': (8, 0x01f, 14, True),
    'root_limit_error_signal_max': (8, 0x021, 14, True),
    'root_limit_fast1_min': (8, 0x023, 14, True),
    'root_limit_fast1_max': (8, 0x025, 14, True),
    'root_limit_fast2_min': (8, 0x027, 14, True),
    'root_limit_fast2_max': (8, 0x029, 14, True),
    'root_pid_setpoint': (8, 0x02b, 14, True),
    'root_pid_kp': (8, 0x02d, 14, True),
    'root_pid_ki': (8, 0x02f, 14, True),
    'root_pid_reset': (8, 0x031, 1, True),
    'root_pid_kd': (8, 0x032, 14, True),
    'root_slow_decimation': (8, 0x034, 5, True),
    'root_control_signal_clr': (8, 0x035, 1, True),
    'root_control_signal_max': (8, 0x036, 25, False),
    'root_control_signal_min': (8, 0x03a, 25, False),
    'root_combined_error_signal_clr': (8, 0x03e, 1, True),
    'root_combined_error_signal_max': (8, 0x03f, 25, False),
    'root_combined_error_signal_min': (8, 0x043, 25, False),
    'scopegen_external_trigger': (6, 0x000, 1, True),
    'scopegen_dac_a_clr': (6, 0x001, 1, True),
    'scopegen_dac_a_max': (6, 0x002, 25, False),
    'scopegen_dac_a_min': (6, 0x006, 25, False),
    'scopegen_dac_b_clr': (6, 0x00a, 1, True),
    'scopegen_dac_b_max': (6, 0x00b, 25, False),
    'scopegen_dac_b_min': (6, 0x00f, 25, False),
    'scopegen_adc_a_sel': (6, 0x013, 4, True),
    'scopegen_adc_b_sel': (6, 0x014, 4, True),
    'slow_pid_setpoint': (2, 0x000, 14, True),
    'slow_pid_kp': (2, 0x002, 14, True),
    'slow_pid_ki': (2, 0x004, 14, True),
    'slow_pid_reset': (2, 0x006, 1, True),
    'slow_pid_kd': (2, 0x007, 14, True),
    'slow_limit_min': (2, 0x009, 14, True),
    'slow_limit_max': (2, 0x00b, 14, True),
    'slow_out_clr': (2, 0x00d, 1, True),
    'slow_out_max': (2, 0x00e, 25, False),
    'slow_out_min': (2, 0x012, 25, False),
    'xadc_temp': (29, 0x000, 12, False),
    'xadc_v': (29, 0x002, 12, False),
    'xadc_a': (29, 0x004, 12, False),
    'xadc_b': (29, 0x006, 12, False),
    'xadc_c': (29, 0x008, 12, False),
    'xadc_d': (29, 0x00a, 12, False),
}
states = ['force', 'di0', 'di1', 'di2', 'di3', 'di4', 'di5', 'di6', 'di7', 'fast_a_x_sat', 'fast_a_x_railed', 'fast_a_y_sat', 'fast_a_y_railed', 'fast_b_x_sat', 'fast_b_x_railed', 'fast_b_y_sat', 'fast_b_y_railed']
signals = ['zero', 'fast_a_x', 'fast_a_y', 'fast_b_x', 'fast_b_y', 'slow_out', 'scopegen_dac_a', 'scopegen_dac_b', 'root_control_signal', 'root_combined_error_signal']



import numpy as np
from PyQt5 import QtGui
from linien.client.widgets import CustomWidget
from linien.client.connection import MHz, Vpp
from linien.client.utils import param2ui


class ModulationAndRampPanel(QtGui.QWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def ready(self):
        self.ids.modulation_frequency.setKeyboardTracking(False)
        self.ids.modulation_frequency.valueChanged.connect(self.change_modulation_frequency)
        self.ids.modulation_amplitude.setKeyboardTracking(False)
        self.ids.modulation_amplitude.valueChanged.connect(self.change_modulation_amplitude)
        self.ids.ramp_speed.currentIndexChanged.connect(self.change_ramp_speed)

    def connection_established(self):
        params = self.app().parameters
        self.control = self.app().control
        self.parameters = params

        param2ui(
            params.modulation_frequency,
            self.ids.modulation_frequency,
            lambda value: value / MHz
        )
        param2ui(
            params.modulation_amplitude,
            self.ids.modulation_amplitude,
            lambda value: value / Vpp
        )
        param2ui(
            params.ramp_speed,
            self.ids.ramp_speed
        )

    def change_modulation_frequency(self):
        self.parameters.modulation_frequency.value = self.ids.modulation_frequency.value() * MHz
        self.control.write_data()

    def change_modulation_amplitude(self):
        self.parameters.modulation_amplitude.value = self.ids.modulation_amplitude.value() * Vpp
        self.control.write_data()

    def change_ramp_speed(self, decimation):
        self.parameters.ramp_speed.value = decimation
        self.control.write_data()


import os
import sys
import signal
import threading

from plumbum import colors

# it may seem odd to include '.', but for some reason this is needed for
# standalone windows executable
sys.path += ['../', os.path.join('..', '..'), '.']

import linien
from linien.client.gui import QTApp


def run_application():
    print('Linien spectroscopy lock version ' + (colors.bold | linien.__version__))
    gui = QTApp()

    # catch ctrl-c and shutdown
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    sys.exit(gui.app.exec_())


if __name__ == '__main__':
    run_application()



import numpy as np
from time import time
from scipy.signal import correlate, resample


LOW_PASS_FILTER = 0
HIGH_PASS_FILTER = 1

FAST_OUT1 = 0
FAST_OUT2 = 1
ANALOG_OUT0 = 2

def downsample(times, values, max_time_diff, max_N=16384):
    last_time = None

    for idx in reversed(range(len(times))):
        current_time = times[idx]
        remove = False

        if last_time is not None:
            if np.abs(last_time - current_time) < max_time_diff / max_N:
                remove = True

        if remove:
            times.pop(idx)
            values.pop(idx)
        else:
            last_time = current_time


def truncate(times, values, max_time_diff):
    while len(values) > 0:
        if time() - times[0] > max_time_diff:
            times.pop(0)
            values.pop(0)
            continue
        break


def update_control_signal_history(history, to_plot, is_locked, max_time_diff):
    if not to_plot:
        return history

    now = time()

    if is_locked:
        history['values'].append(np.mean(to_plot['control_signal']))
        history['times'].append(time())

        if 'slow' in to_plot:
            history['slow_values'].append(to_plot['slow'])
            history['slow_times'].append(time())
    else:
        history['values'] = []
        history['times'] = []
        history['slow_values'] = []
        history['slow_times'] = []

    # truncate
    truncate(history['times'], history['values'], max_time_diff)
    truncate(history['slow_times'], history['slow_values'], max_time_diff)

    # downsample
    downsample(history['times'], history['values'], max_time_diff)
    downsample(history['slow_times'], history['slow_values'], max_time_diff)

    return history


def determine_shift_by_correlation(zoom_factor, reference_signal, error_signal):
    """Compares two spectra and determines the shift by correlation.

    `zoom_factor` is the zoom factor of `error_signal` with respect to
    `reference_signal`, i.e. it states how much reference signal has to be
    magnified in order to show the same region as the new error signal."""
    length = len(error_signal)
    center_idx = int(length / 2)

    # crop the reference signal such that it shows the same region as the new
    # error signal
    idx_shift = int(length * (1 / zoom_factor / 2))
    zoomed_ref = reference_signal[center_idx - idx_shift:center_idx + idx_shift]

    # correlation is slow on red pitaya --> use at maximum 4096 points
    skip_factor = int(len(zoomed_ref) / 4096)
    if skip_factor < 1:
        skip_factor = 1
    zoomed_ref = zoomed_ref[::skip_factor]

    # now sample the error signal down to the same length as the zoomed
    # reference signal
    downsampled_error_signal = resample(error_signal, len(zoomed_ref))

    correlation = correlate(zoomed_ref, downsampled_error_signal)
    """plt.plot(downsampled_error_signal)
    plt.plot(zoomed_ref)
    plt.show()"""


    """plt.plot(correlation)
    plt.show()"""

    shift = np.argmax(correlation)
    shift = (shift - len(zoomed_ref)) / len(zoomed_ref) * 2 / zoom_factor

    return shift, zoomed_ref, downsampled_error_signal


def get_lock_point(error_signal, x0, x1):
    """Calculates parameters for the autolock based on the initial error signal.

    Takes the `error_signal` and two points (`x0` and `x1`) as arguments. The
    points are the points selected by the user, and we know that we want to
    lock between them.
    """
    length = len(error_signal)

    # the data that is between the user-selected bounds
    cropped_data = np.array(error_signal[x0:x1])

    min_idx = np.argmin(cropped_data)
    max_idx = np.argmax(cropped_data)

    # the y value that is between minimum and maximum
    mean_signal = np.mean([cropped_data[min_idx], cropped_data[max_idx]])
    idxs = sorted([min_idx, max_idx])
    slope_data = np.array(cropped_data[idxs[0]:idxs[1]]) - mean_signal

    zero_idx = x0 + np.min(idxs) + np.argmin(np.abs(slope_data))

    # roll the error signal such that the target lock point is exactly in the
    # center
    roll = -int(zero_idx - (length/2))
    rolled_error_signal = np.roll(error_signal, roll)
    # set all the rolled points to zero such that they don't contribute
    # in the correlation
    if roll < 0:
        rolled_error_signal[roll:] = 0
    else:
        rolled_error_signal[:roll] = 0

    target_slope_rising = max_idx > min_idx
    target_zoom = 16384 / (idxs[1] - idxs[0]) / 1.5

    return mean_signal, target_slope_rising, target_zoom, rolled_error_signal


def control_signal_has_correct_amplitude(control_signal, amplitude_target):
    # we ignore some points as sometimes the triggering is not 100%ly
    # correct, i.e. at the beginning or end of the sample we have a glitch
    edge = int(0.4 * len(control_signal))
    control_signal_center = control_signal[edge:-edge]
    control_signal_amplitude = (
        np.max(control_signal_center) - np.min(control_signal_center)
    ) / len(control_signal_center) * len(control_signal) / 16384

    return np.abs(control_signal_amplitude - amplitude_target) / control_signal_amplitude < 0.1


def convert_channel_mixing_value(value):
    if value <= 0:
        a_value = 128
        b_value = 128 + value
    else:
        a_value = 127 - value
        b_value = 128

    return a_value, b_value


def combine_error_signal(error_signals, dual_channel, channel_mixing, chain_factor_width=8):
    if not dual_channel:
        return error_signals[0]

    a_factor, b_factor = convert_channel_mixing_value(channel_mixing)

    return [
        (a_factor * a + b_factor * b) >> chain_factor_width
        for a, b in zip(*error_signals)
    ]


def check_plot_data(is_locked, plot_data):
    if is_locked:
        if 'error_signal' not in plot_data or 'control_signal' not in plot_data:
            return False
    else:
        if 'error_signal_1' not in plot_data:
            return False
    return True


import pickle
import numpy as np
from math import log
from PyQt5 import QtGui, QtWidgets, QtCore

import linien
from linien.common import check_plot_data
from linien.client.utils import param2ui
from linien.client.config import COLORS
from linien.client.widgets import CustomWidget


ZOOM_STEP = .9


class MainWindow(QtGui.QMainWindow, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_ui('main_window.ui')
        self.setWindowTitle('Linien spectroscopy lock %s' % linien.__version__)

    def ready(self):
        def color_to_hex(color):
            result = ''
            for part_idx in range(3):
                result += ('00' + hex(color[part_idx]).lstrip('0x'))[-2:]

            return '#' + result

        set_color = lambda el, color: el.setStyleSheet('color: ' + color_to_hex(COLORS[color]))

        set_color(self.ids.legend_spectrum_1, 'spectroscopy1')
        set_color(self.ids.legend_spectrum_2, 'spectroscopy2')
        set_color(self.ids.legend_spectrum_combined, 'spectroscopy_combined')
        set_color(self.ids.legend_error_signal, 'spectroscopy_combined')
        set_color(self.ids.legend_control_signal, 'control_signal')
        set_color(self.ids.legend_control_signal_history, 'control_signal_history')
        set_color(self.ids.legend_slow_signal_history, 'slow_history')

        self.ids.zoom_slider.valueChanged.connect(self.change_zoom)
        self.ids.go_left_btn.clicked.connect(self.go_left)
        self.ids.go_right_btn.clicked.connect(self.go_right)

    def connection_established(self):
        self.control = self.app.control
        params = self.app.parameters
        self.parameters = params

        param2ui(
            params.ramp_amplitude,
            self.ids.zoom_slider,
            lambda amplitude: int(log(amplitude, ZOOM_STEP))
        )

        def change_manual_navigation_visibility(*args):
            al_running = params.autolock_running.value
            locked = params.lock.value

            self.get_widget('manual_navigation').setVisible(
                not al_running and not locked
            )
            self.get_widget('top_lock_panel').setVisible(locked)

        params.lock.change(change_manual_navigation_visibility)
        params.autolock_running.change(change_manual_navigation_visibility)

        params.to_plot.change(self.update_std)

        params.pid_on_slow_enabled.change(
            lambda v: self.ids.legend_slow_signal_history.setVisible(v)
        )

        self.ids.settings_toolbox.setCurrentIndex(0)

    def go_right(self):
        self.change_center(True)

    def go_left(self):
        self.change_center(False)

    def change_center(self, positive):
        delta_center = self.parameters.ramp_amplitude.value / 10
        if not positive:
            delta_center *= -1
        new_center = self.parameters.center.value + delta_center

        if np.abs(new_center) + self.parameters.ramp_amplitude.value > 1:
            new_center = np.sign(new_center) * (1 - self.parameters.ramp_amplitude.value)

        self.parameters.center.value = new_center
        self.control.write_data()

    def change_zoom(self, zoom):
        self.parameters.ramp_amplitude.value = ZOOM_STEP ** zoom
        self.control.write_data()

    def update_std(self, to_plot):
        if self.parameters.lock.value and to_plot:
            to_plot = pickle.loads(to_plot)
            if to_plot and check_plot_data(True, to_plot):
                error_signal = to_plot.get('error_signal')
                control_signal = to_plot.get('control_signal')
                if error_signal and control_signal:
                    self.ids.error_std.setText('%.2f' % np.std(error_signal))
                    self.ids.control_std.setText('%.2f' % np.std(control_signal))


import json
import pickle
import numpy as np
from os import path
from PyQt5 import QtGui, QtWidgets

from linien.client.utils import param2ui
from linien.client.widgets import CustomWidget


class ViewPanel(QtGui.QWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def ready(self):
        self.ids.autoscale_y_axis.stateChanged.connect(self.autoscale_changed)
        self.ids.y_axis_minimum.valueChanged.connect(self.y_limits_changed)
        self.ids.y_axis_maximum.valueChanged.connect(self.y_limits_changed)

        self.ids.export_select_file.clicked.connect(self.export_select_file)
        self.ids.export_data.clicked.connect(self.export_data)

    def connection_established(self):
        params = self.app().parameters
        self.control = self.app().control
        self.parameters = params

        param2ui(params.autoscale_y, self.ids.autoscale_y_axis)

        def y_axis_limits_changed(value):
            min_, max_ = list(sorted(value))
            self.ids.y_axis_minimum.setValue(min_)
            self.ids.y_axis_maximum.setValue(max_)
        params.y_axis_limits.change(y_axis_limits_changed)

    def autoscale_changed(self):
        self.parameters.autoscale_y.value = int(self.ids.autoscale_y_axis.checkState())

    def y_limits_changed(self):
        self.parameters.y_axis_limits.value = (
            self.ids.y_axis_minimum.value(),
            self.ids.y_axis_maximum.value()
        )

    def export_select_file(self):
        options = QtWidgets.QFileDialog.Options()
        #options |= QtWidgets.QFileDialog.DontUseNativeDialog
        default_ext = '.json'
        fn, _ = QtWidgets.QFileDialog.getSaveFileName(self,"QFileDialog.getSaveFileName()","","JSON (*%s)" % default_ext, options=options)
        if fn:
            if not fn.endswith(default_ext):
                fn = fn + default_ext
            self.export_fn = fn
            self.ids.export_select_file.setText('File selected: %s' % path.split(fn)[-1])
            self.ids.export_data.setEnabled(True)

    def export_data(self):
        fn = self.export_fn
        counter = 0

        while True:
            if counter > 0:
                name, ext = path.splitext(fn)
                fn_with_suffix = name + '-' + str(counter)
                if ext:
                    fn_with_suffix += ext
            else:
                fn_with_suffix = fn

            try:
                open(fn_with_suffix, 'r')
                counter += 1
                continue
            except FileNotFoundError:
                break

        print('export data to', fn_with_suffix)

        with open(fn_with_suffix, 'w') as f:
            data = dict(self.parameters)
            data['to_plot'] = pickle.loads(data['to_plot'])
            json.dump(data, f)


from PyQt5 import QtGui, QtWidgets, QtCore
from threading import Thread
from traceback import print_exc

import linien
from linien.client.config import load_device_data, save_device_data
from linien.client.widgets import CustomWidget
from linien.client.connection import ConnectionThread
from linien.client.dialogs import LoadingDialog, error_dialog, execute_command, \
    question_dialog
from linien.client.ui.new_device_dialog import NewDeviceDialog


class DeviceManager(QtGui.QMainWindow, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_ui('device_manager.ui')
        self.setWindowTitle('Linien spectroscopy lock %s' % linien.__version__)

    def ready(self):
        self.load_device_data(autoload=True)

    def keyPressEvent(self, event):
        key = event.key()
        if key in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
            self.connect()

    def load_device_data(self, autoload=False):
        devices = load_device_data()
        lst = self.ids.deviceList
        lst.clear()

        for device in devices:
            lst.addItem('%s (%s)' % (device['name'], device['host']))

        if autoload and len(devices) == 1:
            self.connect_to_device(devices[0])

    def connect(self):
        devices = load_device_data()

        if not devices:
            return

        self.connect_to_device(devices[self.get_list_index()])

    def connect_to_device(self, device):
        loading_dialog = LoadingDialog(self, device['host'])
        loading_dialog.show()

        aborted = {}
        def was_aborted(*args):
            aborted['aborted'] = True
        loading_dialog.aborted.connect(was_aborted)

        self.t = ConnectionThread(device)

        def connected(conn):
            loading_dialog.hide()
            if not aborted:
                self.app.connected(conn, conn.parameters, conn.control)
        self.t.connected.connect(connected)

        def server_not_installed():
            loading_dialog.hide()
            if not aborted:
                display_question = """The server is not yet installed on the device. Should it be installed?"""
                if question_dialog(self, display_question):
                    self.install_linien_server(device)
        self.t.server_not_installed.connect(server_not_installed)

        def invalid_server_version(remote_version, client_version):
            loading_dialog.hide()
            if not aborted:
                if client_version != 'dev':
                    display_question = \
                        "The server version (%s) does not match the client (%s) version." \
                        "Should the corresponding server version be installed?" \
                        % (remote_version, client_version)
                    if question_dialog(self, display_question):
                        self.install_linien_server(device, version=client_version)
                else:
                    display_error = \
                        "A production version is installed on the RedPitaya, " \
                        "but the client uses a development version. Stop the " \
                        "server and uninstall the version on the RedPitaya using\n" \
                        "    linien_stop_server;\n" \
                        "    pip3 uninstall linien-server\n" \
                        "before trying it again."
                    error_dialog(self, display_error)
        self.t.invalid_server_version.connect(invalid_server_version)

        def authentication_exception():
            loading_dialog.hide()
            if not aborted:
                display_error = 'Error at SSH authentication. ' \
                        'Check username and password and verify that you ' \
                        'don\'t have any offending SSH keys in your known ' \
                        'hosts file.'
                error_dialog(self, display_error)
        self.t.authentication_exception.connect(authentication_exception)

        def general_connection_error():
            loading_dialog.hide()
            if not aborted:
                display_error = "Unable to connect to device."
                error_dialog(self, display_error)
        self.t.general_connection_error.connect(general_connection_error)

        def exception():
            loading_dialog.hide()
            if not aborted:
                display_error = 'Exception occured when connecting to the device.'
                error_dialog(self, display_error)
        self.t.exception.connect(exception)

        def connection_lost():
            error_dialog(self, 'Lost connection to the server!')
            self.app.close()
        self.t.connection_lost.connect(connection_lost)

        self.t.start()

    def install_linien_server(self, device, version=None):
        version_string = ''
        stop_server_command = ''

        if version is not None:
            version_string = '==' + version
            # stop server if another version of linien is installed
            stop_server_command = 'linien_stop_server;'

        self.ssh_command = execute_command(
            self, device['host'], device['username'], device['password'],
            (
                '%s '
                'pip3 install linien-server%s --no-cache-dir; '
                'linien_install_requirements; '
            ) % (stop_server_command, version_string),
            lambda: self.connect_to_device(device)
        )

    def new_device(self):
        self.dialog = NewDeviceDialog()
        self.dialog.setModal(True)
        self.dialog.show()

        def reload_device_data():
            # not very elegant...
            QtCore.QTimer.singleShot(100, self.load_device_data)

        self.dialog.accepted.connect(reload_device_data)

    def get_list_index(self):
        return self.ids.deviceList.currentIndex().row()

    def remove_device(self):
        devices = load_device_data()

        if not devices:
            return

        devices.pop(self.get_list_index())
        save_device_data(devices)
        self.load_device_data()

    def selected_device_changed(self):
        idx = self.get_list_index()

        disable_buttons = True

        if idx >= 0:
            devices = load_device_data()

            if devices:
                disable_buttons = False

        for btn in [self.ids.connectButton, self.ids.removeButton]:
            btn.setEnabled(not disable_buttons)


import setuptools

import linien
assert linien.__version__ != 'dev'

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="linien",
    version=linien.__version__,
    author="Benjamin Wiegand",
    author_email="highwaychile@posteo.de",
    description="Spectroscopy lock application using RedPitaya",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/hermitdemschoenenleben/linien",
    packages=['linien', 'linien.communication', 'linien.client', 'linien.client.ui'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'linien=linien.client.client:run_application'
        ]
    },
    install_requires=[
        'numpy', 'scipy', 'PyQt5>=5.12.2', 'rpyc>=4.1.0', 'appdirs>=1.4.3',
        'pyqtgraph>=0.10.0', 'paramiko>=2.5.0', 'plumbum>=1.6.7', 'uuid'
    ],
    package_data={
        # IMPORTANT: any changes have to be made in client.spec, too
        # (for the standalone installer)
        '': ['*.ui', 'VERSION']
    }
)